"""Basic iembot/nwsbot implementation."""

import copy
import pickle
import random
import re
import traceback
from datetime import timedelta
from io import StringIO
from pathlib import Path
from xml.etree import ElementTree as ET

from pyiem.util import utc
from twisted.application import internet
from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from twisted.mail.smtp import SMTPSenderFactory
from twisted.python import log
from twisted.python.logfile import DailyLogFile
from twisted.words.protocols.jabber import client, jid, xmlstream
from twisted.words.xish import xpath
from twisted.words.xish.domish import Element

from iembot import DATADIR
from iembot.atmosphere import (
    ATManager,
    load_atmosphere_from_db,
)
from iembot.mastodon import load_mastodon_from_db
from iembot.msghandlers import (
    process_groupchat,
    process_privatechat,
)
from iembot.slack import load_slack_from_db
from iembot.twitter import load_twitter_from_db
from iembot.types import JabberClient as JabberClientType
from iembot.util import (
    channels_room_add,
    channels_room_del,
    channels_room_list,
    daily_timestamp,
    email_error,
    load_chatlog,
    load_chatrooms_from_db,
    load_webhooks_from_db,
    purge_logs,
)

# http://stackoverflow.com/questions/7016602
SMTPSenderFactory.noisy = False

PRESENCE_MUC_ITEM = (
    "/presence/x[@xmlns='http://jabber.org/protocol/muc#user']/item"
)
PRESENCE_MUC_STATUS = (
    "/presence/x[@xmlns='http://jabber.org/protocol/muc#user']/status"
)


class JabberClient(JabberClientType):
    """Here lies the Jabber Bot.

    Attributes:
        startup_time (datetime): UTC timestamp when the bot instance started.
        picklefile (str): Path to the persisted chat log pickle file.
        name (str): Bot name / prefix used for database table names.
        dbpool: Database connection pool.
        memcache_client: Optional memcache client instance.
        config (dict): Runtime configuration key/value mapping.
        outstanding_pings (list): Ping tracking list for keepalive logic.
        rooms (dict): Mapping of room name to room metadata.
        chatlog (dict): In-memory chat log storage keyed by room.
        seqnum (int): Latest chat log sequence number.
        routingtable (dict): Channel routing table for room subscriptions.
        at_manager (ATManager): ATmosphere message manager.
        tw_users (dict): Twitter user map keyed by user_id.
        tw_routingtable (dict): Twitter routing map keyed by channel.
        at_users (dict): Atmosphere user map keyed by user_id.
        at_routingtable (dict): Atmosphere routing map keyed by channel.
        md_users (dict): Mastodon user map keyed by user_id.
        md_routingtable (dict): Mastodon routing map keyed by channel.
        slack_teams (dict): Slack teams.
        slack_routingtable (dict): Slack routing map keyed by channel.
        webhooks_routingtable (dict): Webhooks routing map keyed by channel.
        xmlstream: Active XMPP XML stream (if connected).
        firstlogin (bool): Whether the bot has completed first login.
        xmllog (DailyLogFile): XML log file handler.
        myjid: JID representing the bot user.
        ingestjid: JID used for ingest workflows.
        conference (str): Conference (MUC) service hostname.
        email_timestamps (list[datetime]): Timestamps of error emails sent.
        keepalive_lc (LoopingCall | None): Keepalive loop handle.
        fortunes (list[str]): Loaded fortunes used by the bot.
    """

    def __init__(
        self,
        name: str,
        dbpool,
        config=None,
        memcache_client=None,
    ):
        """Constructor"""
        self.startup_time = utc()
        self.picklefile = f"{name}_chatlog_v2.pickle"
        self.name = name
        self.dbpool = dbpool
        self.memcache_client = memcache_client
        self.config = config or {}
        # Adds entries of ping requests made to the server and if we get
        # a response. If this gets to 5 items, we reconnect.
        self.outstanding_pings = []
        self.rooms = {}
        self.chatlog = {}
        self.seqnum = 0
        self.routingtable = {}
        self.at_manager = ATManager()
        self.at_users = {}
        self.at_routingtable = {}
        self.tw_users = {}
        self.tw_routingtable = {}
        # Storage by user_id => {access_token: ..., api_base_url: ...}
        self.md_users = {}
        self.md_routingtable = {}  # Storage by channel => [user_id, ]
        # Slack integration
        self.slack_teams = {}
        self.slack_routingtable = {}
        # Webhooks integration
        self.webhooks_routingtable = {}
        self.xmlstream = None
        self.firstlogin = False
        self.xmllog = DailyLogFile(
            "xmllog", "logs" if Path("logs").is_dir() else "."
        )
        self.myjid = None
        self.ingestjid = None
        self.conference = None
        self.email_timestamps = []
        self.keepalive_lc = None  # Keepalive LoopingCall
        with open(DATADIR / "startrek", encoding="utf-8") as fp:
            self.fortunes = fp.read().split("\n%\n")
        load_chatlog(self)

        lc2 = LoopingCall(purge_logs, self)
        lc2.start(60 * 60 * 24)
        lc3 = LoopingCall(reactor.callInThread, self.save_chatlog)
        lc3.start(600)  # Every 10 minutes

    def save_chatlog(self):
        """called from a thread"""
        log.msg(f"Saving CHATLOG to {self.picklefile}")
        with open(self.picklefile, "wb") as fh:
            # unsure if deepcopy is necessary, but alas
            pickle.dump(copy.deepcopy(self.chatlog), fh)

    def reload_config(self, always_join: bool):
        """Jobs the bot should do when a database reload is needed."""
        self.load_twitter()
        self.load_mastodon()
        self.load_chatrooms(always_join)
        self.load_webhooks()
        self.load_slack()
        self.load_atmosphere()

    def authd(self, _xs=None):
        """callback when we are logged into the server!"""
        email_error(None, self, f"Logged into jabber server as {self.myjid}")
        if not self.firstlogin:
            self.compute_daily_caller()
            self.firstlogin = True

        # Resets associated with the previous login session, perhaps
        self.rooms = {}
        self.outstanding_pings = []

        self.reload_config(always_join=True)

        # Start the keepalive loop
        if self.keepalive_lc is None:
            self.keepalive_lc = LoopingCall(self.housekeeping)
            self.keepalive_lc.start(60)

    def next_seqnum(self):
        """
        Simple tool to generate a sequence number for message logging
        """
        self.seqnum += 1
        return self.seqnum

    def load_chatrooms(self, always_join: bool):
        """
        Load up the chatrooms and subscriptions from the database!, I also
        support getting called at a later date for any changes
        """
        log.msg("load_chatrooms() called...")
        df = self.dbpool.runInteraction(
            load_chatrooms_from_db, self, always_join
        )
        # Send a presence update, which in the case of the first login will
        # provoke any offline messages to be sent.
        df.addCallback(self.send_presence)
        df.addErrback(email_error, self, "load_chatrooms() failure")

    def load_twitter(self):
        """Load the twitter subscriptions and access tokens"""
        log.msg("load_twitter() called...")
        df = self.dbpool.runInteraction(load_twitter_from_db, self)
        df.addErrback(email_error, self, "load_twitter() failure")

    def load_atmosphere(self):
        """Load the atmosphere subscriptions and access tokens"""
        log.msg("load_atmosphere() called...")
        df = self.dbpool.runInteraction(load_atmosphere_from_db, self)
        df.addErrback(email_error, self, "load_atmosphere() failure")

    def load_slack(self):
        """Load the slack subscriptions and access tokens"""
        log.msg("load_slack() called...")
        df = self.dbpool.runInteraction(load_slack_from_db, self)
        df.addErrback(email_error, self, "load_slack() failure")

    def load_mastodon(self):
        """Load the Mastodon subscriptions and access tokens"""
        log.msg("load_mastodon() called...")
        df = self.dbpool.runInteraction(load_mastodon_from_db, self)
        df.addErrback(email_error, self, "load_mastodon() failure")

    def load_webhooks(self):
        """Load the twitter subscriptions and access tokens"""
        log.msg("load_webhooks() called...")
        df = self.dbpool.runInteraction(load_webhooks_from_db, self)
        df.addErrback(email_error, self, "load_webhooks() failure")

    def fire_client(self, _res, serviceCollection):
        """Callback from main saying that our database (select 1) is ready."""
        log.msg("fire_client() called ...")

        self.myjid = jid.JID(
            f"{self.config.get('bot.username', 'iembot')}@"
            f"{self.config.get('bot.xmppdomain', 'localhost')}/"
            "twisted_words"
        )
        self.ingestjid = jid.JID(
            f"{self.config.get('bot.ingest_username', 'iembot-ingest')}@"
            f"{self.config.get('bot.xmppdomain', 'localhost')}"
        )
        self.conference = self.config.get(
            "bot.mucservice", "conference.localhost"
        )

        factory = client.XMPPClientFactory(
            self.myjid, self.config.get("bot.password", "")
        )
        # Limit reconnection delay to 60 seconds
        factory.maxDelay = 60
        factory.addBootstrap(xmlstream.STREAM_CONNECTED_EVENT, self.connected)
        factory.addBootstrap(xmlstream.STREAM_AUTHD_EVENT, self.authd)
        factory.addBootstrap(xmlstream.INIT_FAILED_EVENT, log.err)
        factory.addBootstrap(xmlstream.STREAM_END_EVENT, self.disconnected)

        # pylint: disable=no-member
        i = internet.TCPClient(
            self.config.get("bot.connecthost", "127.0.0.1"), 5222, factory
        )
        i.setServiceParent(serviceCollection)

    def connected(self, xs):
        """connected callback"""
        log.msg("Connected")
        self.xmlstream = xs
        self.xmlstream.rawDataInFn = self.rawDataInFn
        self.xmlstream.rawDataOutFn = self.rawDataOutFn

        self.xmlstream.addObserver("/message", self.on_message)
        self.xmlstream.addObserver("/iq", self.on_iq)
        self.xmlstream.addObserver("/presence/x/item", self.on_presence)

    def disconnected(self, _xs=None):
        """disconnected callback"""
        log.msg("disconnected() was called...")

    def get_fortune(self):
        """Get a random value from the array"""
        offset = int((len(self.fortunes) - 1) * random.random())
        return " ".join(self.fortunes[offset].replace("\n", "").split())

    def rawDataInFn(self, data):
        """write xmllog"""
        self.xmllog.write(
            f"{utc():%Y-%m-%d %H:%M:%S} RECV "
            f"{data.decode('utf-8', 'ignore')}\n"
        )

    def rawDataOutFn(self, data):
        """write xmllog"""
        self.xmllog.write(
            f"{utc():%Y-%m-%d %H:%M:%S} SEND "
            f"{data.decode('utf-8', 'ignore')}\n"
        )

    def housekeeping(self):
        """
        This gets exec'd every minute to keep up after ourselves
        1. XMPP Server Ping
        2. Update presence
        """
        if self.outstanding_pings:
            log.msg(f"Currently unresponded pings: {self.outstanding_pings}")
        if len(self.outstanding_pings) > 5:
            self.outstanding_pings = []
            email_error(
                "Logging out of Chat!", self, "IQ error limit reached..."
            )
            if self.xmlstream is not None:
                # send a disconnect
                self.xmlstream.sendFooter()
            return
        if self.xmlstream is None or self.myjid is None:
            log.msg("xmlstream or myjid is None, not sending ping")
            return
        utcnow = utc()
        ping = Element((None, "iq"))
        ping["to"] = self.myjid.host
        ping["type"] = "get"
        pingid = f"{utcnow:%Y%m%d%H%M.%S}"
        ping["id"] = pingid
        ping.addChild(Element(("urn:xmpp:ping", "ping")))
        self.outstanding_pings.append(pingid)
        self.xmlstream.send(ping)
        # Update our presence every ten minutes with some debugging info
        if utcnow.minute % 10 == 0:
            self.send_presence()

    def send_privatechat(self, to, mess, htmlstr=None):
        """
        Helper method to send private messages

        @param to: String jabber ID to send the message too
        @param mess: String plain text version to send
        @param html: Optional String html version
        """
        message = Element(("jabber:client", "message"))
        if to.find("@") == -1:  # base username, add domain
            to = f"{to}@{self.config['bot.xmppdomain']}"
        message["to"] = to
        message["type"] = "chat"
        message.addElement("body", None, mess)
        html = message.addElement(
            "html", "http://jabber.org/protocol/xhtml-im"
        )
        body = html.addElement("body", "http://www.w3.org/1999/xhtml")
        if htmlstr is not None:
            body.addRawXml(htmlstr)
        else:
            p = body.addElement("p")
            p.addContent(mess)
        self.xmlstream.send(message)

    def send_groupchat(self, room, plain, htmlstr=None):
        """Send a groupchat message to a given room

        Args:
          room (str): The roomname (which we should have already joined)
          plain (str): The message to send to the room, no escaping necessary
          htmlstr (str, optional): The HTML variant of the message

        Returns:
          twisted.words.xish.domish.Element: the element that was sent
        """
        message = Element(("jabber:client", "message"))
        message["to"] = f"{room}@{self.conference}"
        message["type"] = "groupchat"
        message.addElement("body", None, plain)
        html = message.addElement(
            "html", "http://jabber.org/protocol/xhtml-im"
        )
        body = html.addElement("body", "http://www.w3.org/1999/xhtml")
        if htmlstr is not None:
            body.addRawXml(htmlstr)
        else:
            # Careful here, we always want to have valid xhtml, so we should
            # wrap plain text in a paragraph tag
            p = body.addElement("p")
            p.addContent(plain)
        # Ensure that we have well formed XML before sending it
        try:
            ET.fromstring(message.toXml())
        except Exception as exp:
            email_error(exp, self, message.toXml())
            return None
        else:
            self.send_groupchat_elem(message)
        return message

    def send_groupchat_elem(self, elem, to=None, secondtrip=False):
        """Wrapper for sending groupchat elements"""
        if to is not None:
            elem["to"] = to
        room = jid.JID(elem["to"]).user
        if room not in self.rooms:
            email_error(
                f"Attempted to send message to room [{room}] "
                "we have not joined...",
                self,
                elem,
            )
            return
        if not self.rooms[room]["joined"]:
            if secondtrip:
                log.msg(
                    f"ABORT of send to room: {room}, msg: {elem}, not in room"
                )
                return
            secs = random.randint(0, 10)
            log.msg(f"delaying by {secs}s send to: {room}, not in room yet")
            # Need to prevent elem['to'] object overwriting
            reactor.callLater(secs, self.send_groupchat_elem, elem, elem["to"])
            return
        self.xmlstream.send(elem)

    def send_presence(self, _=None):
        """
        Set a presence for my login, could be from a callback (load_chatrooms).
        """
        presence = Element(("jabber:client", "presence"))
        msg = (
            f"Booted: {self.startup_time:%d %b} "
            f"Updated: {utc():%H%M} UTC, Rooms: {len(self.rooms)}, "
            f"Messages: {self.seqnum}"
        )
        presence.addElement("status").addContent(msg)
        if self.xmlstream is not None:
            self.xmlstream.send(presence)

    def compute_daily_caller(self):
        """Figure out when to be called"""
        log.msg("compute_daily_caller() called...")
        # Figure out when to spam all rooms with a timestamp
        utcnow = utc() + timedelta(days=1)
        tnext = utcnow.replace(hour=0, minute=0, second=0)
        log.msg(
            "Initial Calling daily_timestamp in "
            f"{(tnext - utc()).seconds} seconds"
        )
        reactor.callLater((tnext - utc()).seconds, daily_timestamp, self)

    def presence_processor(self, elem):
        """Process the presence stanza

                The bot process keeps track of room occupants and their
                affiliations, roles so to provide ACLs for room admin
                activities.

                Args:
                  elem (domish.Element): stanza

        <presence xmlns='jabber:client' to='nwsbot@laptop.local/twisted_words'
            from='dmxchat@conference.laptop.local/nws-daryl.herzmann'>
            <priority>1</priority>
            <c xmlns='http://jabber.org/protocol/caps'
            node='http://pidgin.im/'
                ver='AcN1/PEN8nq7AHD+9jpxMV4U6YM='
                ext='voice-v1 camera-v1 video-v1'
                hash='sha-1'/>
            <x xmlns='vcard-temp:x:update'><photo/></x>
            <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='owner' jid='nws-mortal@laptop.local/laptop'
                role='moderator'/>
            </x>
        </presence>
        """
        items = xpath.queryForNodes(PRESENCE_MUC_ITEM, elem)
        if items is None:
            return

        _room = jid.JID(elem["from"]).user
        if _room not in self.rooms:
            email_error(
                f"Got MUC presence from unknown room '{_room}'",
                self,
                elem,
            )
            return
        _handle = jid.JID(elem["from"]).resource
        statuses = xpath.queryForNodes(PRESENCE_MUC_STATUS, elem)
        muc_codes = []
        if statuses is not None:
            muc_codes = [status.getAttribute("code") for status in statuses]

        for item in items:
            affiliation = item.getAttribute("affiliation")
            _jid = item.getAttribute("jid")
            role = item.getAttribute("role")
            left = affiliation == "none" and role == "none"
            selfpres = "110" in muc_codes
            if selfpres:
                log.msg(f"MUC '{_room}' self presence left: {left}")
                self.rooms[_room]["joined"] = not left

            self.rooms[_room]["occupants"][_handle] = {
                "jid": _jid,
                "affiliation": affiliation,
                "role": role,
            }

    def iq_processor(self, elem: Element):
        """Response to IQ stanzas."""
        typ = elem.getAttribute("type")
        # A response is being requested of us.
        if typ == "get" and elem.firstChildElement().name == "ping":
            # Respond to a ping request.
            pong = Element((None, "iq"))
            pong["type"] = "result"
            pong["to"] = elem["from"]
            pong["from"] = elem["to"]
            pong["id"] = elem["id"]
            self.xmlstream.send(pong)
        # We are getting a response to a request we sent, maybe.
        elif typ == "result":
            if elem.getAttribute("id") in self.outstanding_pings:
                self.outstanding_pings.remove(elem.getAttribute("id"))

    def send_help_message(self, user):
        """
        Send a user a help message about what I can do
        """
        me = self.myjid.user
        msg = (
            f"Hi, I am {me}.  You can try talking directly "
            "with me. I currently do not support any commands, sorry."
        )
        htmlmsg = msg.replace("\n", "<br />").replace(
            self.myjid.user,
            f'<a href="https://{self.myjid.host}/{me}faq.php">%s</a>',
        )
        self.send_privatechat(user, msg, htmlmsg)

    def convert_to_privatechat(self, myjid):
        """
        The bot can't handle private chats via a MUC, readdress to a
        private chat

        @param myjid: MUC chat jabber ID to reroute
        """
        _handle = myjid.resource
        _room = myjid.user
        if _handle not in self.rooms[_room]["occupants"]:
            return
        realjid = self.rooms[_room]["occupants"][_handle]["jid"]

        self.send_help_message(realjid)

        message = Element(("jabber:client", "message"))
        message["to"] = myjid.full()
        message["type"] = "chat"
        message.addElement(
            "body",
            None,
            (
                "I can't help you here, please chat "
                "with me outside of a groupchat.  "
                "I have initated such a chat for "
                "you."
            ),
        )
        self.xmlstream.send(message)

    def message_processor(self, elem: Element):
        """
        This is our business method, figure out if this chat is a
        private message or a group one
        """
        if elem.hasAttribute("type") and elem["type"] == "groupchat":
            process_groupchat(self, elem)
        elif elem.hasAttribute("type") and elem["type"] == "error":
            email_error("Got Error Stanza?", self, elem.toXml())
        else:
            process_privatechat(self, elem)

    def on_message(self, elem: Element):
        """We got a message!"""
        self.stanza_processor(self.message_processor, elem)

    def on_presence(self, elem: Element):
        """We got a presence"""
        self.stanza_processor(self.presence_processor, elem)

    def on_iq(self, elem: Element):
        """We got an IQ"""
        self.stanza_processor(self.iq_processor, elem)

    def stanza_processor(self, func: callable, elem: Element):
        """Process the given stanza on the main thread to ensure order.

        Args:
            func (callable): Function to call with the stanza
            elem (domish.Element): Stanza to process
        """
        try:
            func(elem)
        except Exception as exp:
            log.err(exp)
            io = StringIO()
            traceback.print_exc(file=io)
            email_error(io.getvalue(), self, elem.toXml())

    def talkWithUser(self, elem: Element):
        """
        Look for commands that a user may be asking me for
        @param elem domish.Element to process
        """
        if not elem.body:
            log.msg("Empty conversation?")
            log.msg(elem.toXml())
            return

        bstring = elem.body.lower()
        if re.match(r"^flood", bstring):
            self.handle_flood_request(elem, bstring)
        else:
            self.send_help_message(elem["from"])

    def process_groupchat_cmd(self, room, res, cmd):
        """
        Logic to process any groupchat commands proferred to nwsbot

        @param room String roomname that this command came from
        @param res String value of the resource that sent the command
        @param cmd String command that the resource sent
        """
        # Make sure we know who the real JID of this user is....
        if res not in self.rooms[room]["occupants"]:
            self.send_groupchat(
                room,
                (
                    f"{res}: Sorry, I am unable to process "
                    "your request due to a lookup failure."
                    "  Please consider rejoining the "
                    "chatroom if you really wish to "
                    "speak with me."
                ),
            )
            return

        # Figure out the user's affiliation
        aff = self.rooms[room]["occupants"][res]["affiliation"]
        _jid = self.rooms[room]["occupants"][res]["jid"]

        # Support legacy ping, return as done
        if re.match(r"^ping", cmd, re.IGNORECASE):
            self.send_groupchat(room, f"{res}: pong")

        # Listing of channels is not admin privs
        elif re.match(r"^channels list", cmd, re.IGNORECASE):
            channels_room_list(self, room)

        # Add a channel to the room's subscriptions
        elif re.match(r"^channels add", cmd, re.IGNORECASE):
            add_channel = cmd[12:].strip().upper()
            if aff in ["owner", "admin"]:
                if len(add_channel) < 24:
                    df = self.dbpool.runInteraction(
                        channels_room_add, self, room, cmd[12:]
                    )
                    df.addErrback(email_error, self, room + " -> " + cmd)
                else:
                    err = (
                        f"{res}: Error, channels are less than 24 characters!"
                    )
                    self.send_groupchat(room, err)
            else:
                err = (
                    f"{res}: Sorry, you must be a room admin to add a channel"
                )
                self.send_groupchat(room, err)

        # Del a channel to the room's subscriptions
        elif re.match(r"^channels del", cmd, re.IGNORECASE):
            if aff in ["owner", "admin"]:
                df = self.dbpool.runInteraction(
                    channels_room_del, self, room, cmd[12:]
                )
                df.addErrback(email_error, self, f"{room} -> {cmd}")
            else:
                err = (
                    f"{res}: Sorry, you must be a room admin to add a channel"
                )
                self.send_groupchat(room, err)

        # Look for users request
        elif re.match(r"^users", cmd, re.IGNORECASE):
            if _jid is None:
                err = "Sorry, I am not able to see room occupants."
                self.send_groupchat(room, err)
            elif aff in ["owner", "admin"]:
                rmess = ""
                for hndle in self.rooms[room]["occupants"].keys():
                    rmess += (
                        f"{hndle} "
                        f"({self.rooms[room]['occupants'][hndle]['jid']}), "
                    )
                self.send_privatechat(_jid, f"JIDs in room: {rmess}")
            else:
                err = f"{res}: Sorry, you must be a room admin to query users"
                self.send_groupchat(room, err)

        # Else send error message about what I support
        else:
            err = f"ERROR: unsupported command: '{cmd}'"
            self.send_groupchat(room, err)
            self.send_groupchat_help(room)

    def send_groupchat_help(self, room):
        """
        Send a message to a given chatroom about what commands I support
        """
        msg = (
            "Current Supported Commands:\n"
            f"{self.myjid.user}: channels add channelname[,channelname] "
            "### Add channel subscription(s) for this room\n"
            f"{self.myjid.user}: channels del channelname[,channelname] "
            "### Delete channel subscriptions(s) for this room\n"
            f"{self.myjid.user}: channels list "
            "### List channels this room is subscribed to\n"
            f"{self.myjid.user}: ping          "
            "### Test connectivity with a 'pong' response\n"
            f"{self.myjid.user}: users   ### Generates list of users in room"
        )

        htmlmsg = msg.replace("\n", "<br />")
        self.send_groupchat(room, msg, htmlmsg)

    def handle_flood_request(self, elem, bstring):
        """
        All a user to flood a chatroom with messages to flush it!
        with star trek quotes, yes!
        """
        _from = jid.JID(elem["from"])
        if not re.match("^nws-", str(_from.user)):
            msg = "Sorry, you must be NWS to flood a chatroom!"
            self.send_privatechat(elem["from"], msg)
            return
        tokens = bstring.split()
        if len(tokens) == 1:
            msg = "Did you specify a room to flood?"
            self.send_privatechat(elem["from"], msg)
            return
        room = tokens[1].lower()
        for _i in range(60):
            self.send_groupchat(room, self.get_fortune())
        self.send_groupchat(
            room,
            (
                "Room flooding complete, offending message "
                "should no longer appear"
            ),
        )



"""Mastodon stuff."""

import time

import mastodon as Mastodon
import requests
from mastodon.errors import MastodonUnauthorizedError
from twisted.internet import threads
from twisted.python import log
from twisted.words.xish.domish import Element

from iembot.twitter import safe_twitter_text
from iembot.types import JabberClient
from iembot.util import build_channel_subs, email_error


def load_mastodon_from_db(txn, bot: JabberClient):
    """Load Mastodon config from database"""

    bot.md_routingtable = build_channel_subs(
        txn,
        "iembot_mastodon_oauth",
    )

    mdusers = {}
    txn.execute(
        """
        select server, o.iembot_account_id,
        o.access_token, o.screen_name, o.iem_owned
        from iembot_mastodon_apps a JOIN iembot_mastodon_oauth o
            on (a.id = o.appid) WHERE o.access_token is not null and
        not o.disabled
        """
    )
    for row in txn.fetchall():
        mdusers[row["iembot_account_id"]] = {
            "screen_name": row["screen_name"],
            "access_token": row["access_token"],
            "api_base_url": row["server"],
            "iem_owned": row["iem_owned"],
        }
    bot.md_users = mdusers
    log.msg(f"load_mastodon_from_db(): {txn.rowcount} access tokens found")


def disable_mastodon_user(bot: JabberClient, user_id, errcode=0):
    """Disable the Mastodon subs for this user_id

    Args:
        user_id (big_id): The Mastodon user to disable
        errcode (int): The HTTP-like errorcode
    """
    mduser = bot.md_users.get(user_id)
    if mduser is None:
        log.msg(f"Failed to disable unknown Mastodon user_id {user_id}")
        return False
    screen_name = mduser["screen_name"]
    if mduser["iem_owned"]:
        log.msg(
            f"Skipping disabling of Mastodon for {user_id} ({screen_name})"
        )
        return False
    bot.md_users.pop(user_id, None)
    log.msg(
        f"Removing Mastodon access token for user: {user_id} ({screen_name}) "
        f"errcode: {errcode}"
    )
    df = bot.dbpool.runOperation(
        "UPDATE iembot_mastodon_oauth SET updated = now(), "
        "access_token = null, api_base_url = null "
        "WHERE user_id = %s",
        (user_id,),
    )
    df.addErrback(log.err)
    return True


def toot_cb(response, bot: JabberClient, twttxt, user_id):
    """
    Called after success going to Mastodon
    """
    if response is None:
        return None
    mduser = bot.md_users.get(user_id)
    if mduser is None:
        return response
    if "content" not in response:
        log.msg(f"Got response without content {response}")
        return None
    url = response["url"]

    response.pop(
        "account", None
    )  # Remove extra junk, there's still a lot more though...

    # Log
    df = bot.dbpool.runOperation(
        "INSERT into iembot_social_log(medium, source, resource_uri, "
        "message, response, response_code, iembot_account_id) "
        "values (%s,%s,%s,%s,%s,%s,%s)",
        ("mastodon", "", url, twttxt, repr(response), 200, user_id),
    )
    df.addErrback(log.err)
    return response


def mastodon_errback(err, bot: JabberClient, user_id, tweettext):
    """Error callback when simple Mastodon workflow fails."""
    # Always log it
    log.err(err)
    errcode = None
    if isinstance(err, Mastodon.errors.MastodonNotFoundError):
        errcode = 404
        disable_mastodon_user(bot, user_id, errcode)
    elif isinstance(err, Mastodon.errors.MastodonUnauthorizedError):
        errcode = 401
        disable_mastodon_user(bot, user_id, errcode)
    else:
        sn = bot.md_users.get(user_id, {}).get("screen_name", "")
        msg = f"User: {user_id} ({sn})\nFailed to toot: {tweettext}"
        email_error(err, bot, msg)


def toot(self, user_id, twttxt, **kwargs):
    """
    Send a message to Mastodon
    """
    twttxt = safe_twitter_text(twttxt)
    df = threads.deferToThread(
        really_toot,
        self,
        user_id,
        twttxt,
        **kwargs,
    )
    df.addCallback(toot_cb, self, twttxt, user_id)
    df.addErrback(
        mastodon_errback,
        self,
        user_id,
        twttxt,
    )
    df.addErrback(
        email_error,
        self,
        f"User: {user_id}, Text: {twttxt} Hit double exception",
    )
    return df


def really_toot(bot: JabberClient, user_id, twttxt, **kwargs):
    """Blocking Mastodon toot method."""
    if user_id not in bot.md_users:
        log.msg(f"toot() called with unknown Mastodon user_id: {user_id}")
        return None
    api = Mastodon.Mastodon(
        access_token=bot.md_users[user_id]["access_token"],
        api_base_url=bot.md_users[user_id]["api_base_url"],
    )
    log.msg(
        "Sending to Mastodon "
        f"{bot.md_users[user_id]['screen_name']}({user_id}) "
        f"'{twttxt}' media:{kwargs.get('twitter_media')}"
    )
    media = kwargs.get("twitter_media")
    media_id = None

    res = None
    try:
        params = {
            "status": twttxt,
        }
        # If we have media, we have some work to do!
        if media is not None:
            resp = requests.get(media, timeout=30, stream=True)
            resp.raise_for_status()
            # TODO: Is this always image/png?
            media_id = api.media_post(resp.raw, mime_type="image/png")
            params["media_ids"] = [media_id]
        res = api.status_post(**params)
    except MastodonUnauthorizedError:
        # Access token is no longer valid
        disable_mastodon_user(bot, user_id)
        return None
    except Mastodon.errors.MastodonRatelimitError as exp:
        # Submitted too quickly
        log.err(exp)
        # Since this called from a thread, sleeping should not jam us up
        time.sleep(kwargs.get("sleep", 10))
        res = api.status_post(**params)
    except Mastodon.errors.MastodonError as exp:
        # Something else bad happened when submitting this to the Mastodon
        log.err(exp)
        params.pop("media_ids", None)  # Try again without media
        # Since this called from a thread, sleeping should not jam us up
        time.sleep(kwargs.get("sleep", 10))
        try:
            res = api.status_post(**params)
        except Mastodon.errors.MastodonError as exp2:
            log.err(exp2)
    except Exception as exp:
        # Something beyond Mastodon went wrong
        log.err(exp)
        params.pop("media_ids", None)  # Try again without media
        # Since this called from a thread, sleeping should not jam us up
        time.sleep(kwargs.get("sleep", 10))
        res = api.status_post(**params)
    return res


def route(bot: JabberClient, channels: list, elem: Element):
    """Do Maston stuff."""
    alertedPages = []
    for channel in channels:
        for user_id in bot.md_routingtable.get(channel, []):
            if user_id not in bot.md_users:
                log.msg(
                    "Failed to send to Mastodon due to no "
                    f"access_tokens {user_id}"
                )
                continue
            # Require the x.twitter attribute to be set to prevent
            # confusion with some ingestors still sending tweets themselfs
            if not elem.x.hasAttribute("twitter"):
                continue
            if user_id in alertedPages:
                continue
            alertedPages.append(user_id)
            lat = long = None
            if (
                elem.x
                and elem.x.hasAttribute("lat")
                and elem.x.hasAttribute("long")
            ):
                lat = elem.x["lat"]
                long = elem.x["long"]
            really_toot(
                bot,
                user_id,
                elem.x["twitter"],
                twitter_media=elem.x.getAttribute("twitter_media"),
                latitude=lat,  # TODO: unused
                longitude=long,  # TODO: unused
            )



"""Logic handling messages the bot receives."""

from pyiem.util import utc
from twisted.internet import reactor
from twisted.python import log
from twisted.words.protocols.jabber import jid
from twisted.words.xish import xpath
from twisted.words.xish.domish import Element

from iembot import ROOM_LOG_ENTRY
from iembot.types import JabberClient

REGISTERED_HANDLERS = []


def register_handler(handler: callable):
    """Add a message handler to the registry."""
    if handler not in REGISTERED_HANDLERS:
        REGISTERED_HANDLERS.append(handler)


def process_privatechat(bot: JabberClient, elem: Element) -> None:
    """Starting point for processing a private (1 on 1) chat.

    Args:
      bot (JabberClient): The running bot instance receiving the message
      elem (Element): The XML Received.
    """
    _from = jid.JID(elem["from"])
    if elem["from"] == bot.config["bot.xmppdomain"]:
        log.msg("MESSAGE FROM SERVER?")
        return
    # Intercept private messages via a chatroom, can't do that :)
    if _from.host == bot.config["bot.mucservice"]:
        log.msg("ERROR: message is MUC private chat")
        return

    if _from.userhost() != f"iembot_ingest@{bot.config['bot.xmppdomain']}":
        log.msg("ERROR: message not from iembot_ingest")
        return

    # Ready for launch
    process_message_from_ingest(bot, elem)


def process_message_from_ingest(bot: JabberClient, elem: Element) -> None:
    """Process a message received from the ingest system."""

    # Go look for body to see routing info!
    # Get the body string
    bstring = xpath.queryForString("/message/body", elem)
    if not bstring:
        log.msg("Nothing found in body?")
        return

    if elem.x and elem.x.hasAttribute("channels"):
        channels = elem.x["channels"].split(",")
    else:
        # The body string contains
        channel = bstring.split(":", 1)[0]
        channels = [channel]

    # Always send to botstalk
    elem["to"] = f"botstalk@{bot.config['bot.mucservice']}"
    elem["type"] = "groupchat"
    bot.send_groupchat_elem(elem)

    for handler in REGISTERED_HANDLERS:
        handler(bot, channels, elem)


def process_groupchat(bot: JabberClient, elem: Element) -> None:
    """Starting point for groupchat processing.

    Args:
      bot (JabberClient): Running bot instance
      elem (Element): Payload received.
    """
    # Ignore all messages that are x-stamp (delayed / room history)
    # <delay xmlns='urn:xmpp:delay' stamp='2016-05-06T20:04:17.513Z'
    #  from='nwsbot@laptop.local/twisted_words'/>
    if xpath.queryForNodes("/message/delay[@xmlns='urn:xmpp:delay']", elem):
        return

    _from = jid.JID(elem["from"])
    room: str = _from.user
    res = _from.resource

    body = xpath.queryForString("/message/body", elem)
    if body is not None and len(body) >= 4 and body[:4] == "ping":
        bot.send_groupchat(room, f"{res}: {bot.get_fortune()}")

    # Look for bot commands
    if body.startswith(bot.name):
        bot.process_groupchat_cmd(room, res, body[7:].strip())

    # In order for the message to be logged, it needs to be from iembot
    # and have a channels attribute
    if res is None or res != "iembot":
        return

    a = xpath.queryForNodes("/message/x[@xmlns='nwschat:nwsbot']", elem)
    if a is None or not a:
        return

    roomlog = bot.chatlog.setdefault(room, [])
    ts = utc()

    product_id = ""
    if elem.x and elem.x.hasAttribute("product_id"):
        product_id = elem.x["product_id"]

    html = xpath.queryForNodes("/message/html/body", elem)
    log_entry = body
    if html is not None:
        log_entry = html[0].toXml()

    if len(roomlog) > 40:
        roomlog.pop()

    def writelog(product_text=None):
        """Actually do what we want to do"""
        if product_text is None or product_text == "":
            product_text = "Sorry, product text is unavailable."
        roomlog.insert(
            0,
            ROOM_LOG_ENTRY(
                seqnum=bot.next_seqnum(),
                timestamp=ts.strftime("%Y%m%d%H%M%S"),
                log=log_entry,
                author=res,
                product_id=product_id,
                product_text=product_text,
                txtlog=body,
            ),
        )

    if product_id == "":
        writelog()
        return

    def got_data(res, trip):
        """got a response!"""
        (_flag, data) = res
        if data is None:
            if trip < 5:
                reactor.callLater(10, memcache_fetch, trip)
            else:
                writelog()
            return
        if trip > 1:
            log.msg(f"memcache lookup of {product_id} succeeded")
        writelog(data.decode("ascii", "ignore"))

    def no_data(mixed):
        """got no data"""
        log.err(mixed)
        writelog()

    def memcache_fetch(trip):
        """fetch please"""
        trip += 1
        if trip > 1:
            log.msg(f"memcache_fetch(trip={trip}, product_id={product_id}")
        defer = bot.memcache_client.get(product_id.encode("utf-8"))
        defer.addCallback(got_data, trip)
        defer.addErrback(no_data)

    memcache_fetch(0)



"""Twitter/X stuff."""

import json
import re
import time
from html import unescape

import requests
from pyiem.reference import TWEET_CHARS
from requests_oauthlib import OAuth1, OAuth1Session
from twisted.internet import threads
from twisted.python import log
from twisted.words.xish.domish import Element
from twitter.api import Api as TwitterApi
from twitter.error import TwitterError

from iembot.types import JabberClient
from iembot.util import build_channel_subs, email_error

TWEET_API = "https://api.x.com/2/tweets"
# 89: Expired token, so we shall revoke for now
# 185: User is over quota
# 226: Twitter thinks this tweeting user is spammy, le sigh
# 326: User is temporarily locked out
# 64: User is suspended
DISABLE_TWITTER_CODES = [89, 185, 226, 326, 64]


def safe_twitter_text(text: str) -> str:
    """Attempt to rip apart a message that is too long!
    To be safe, the URL is counted as 24 chars
    """
    # XMPP payload will have entities, unescape those before tweeting
    text = unescape(text)
    # Convert two or more spaces into one
    text = " ".join(text.split())
    # If we are already below TWEET_CHARS, we don't have any more work to do...
    if len(text) < TWEET_CHARS and text.find("http") == -1:
        return text
    chars = 0
    words = text.split()
    # URLs only count as 25 chars, so implement better accounting
    for word in words:
        if word.startswith("http"):
            chars += 25
        else:
            chars += len(word) + 1
    if chars < TWEET_CHARS:
        return text
    urls = re.findall(r"https?://[^\s]+", text)
    if len(urls) == 1:
        text2 = text.replace(urls[0], "")
        sections = re.findall("(.*) for (.*)( till [0-9A-Z].*)", text2)
        if len(sections) == 1:
            text = f"{sections[0][0]}{sections[0][2]}{urls[0]}"
            if len(text) > TWEET_CHARS:
                sz = TWEET_CHARS - 26 - len(sections[0][2])
                text = f"{sections[0][0][:sz]}{sections[0][2]}{urls[0]}"
            return text
        if len(text) > TWEET_CHARS:
            # 25 for URL, three dots and space for 29
            return f"{text2[: (TWEET_CHARS - 29)]}... {urls[0]}"
    if chars > TWEET_CHARS and words[-1].startswith("http"):
        i = -2
        while len(" ".join(words[:i])) > (TWEET_CHARS - 3 - 25):
            i -= 1
        return f"{' '.join(words[:i])}... {words[-1]}"
    return text[:TWEET_CHARS]


def load_twitter_from_db(txn, bot: JabberClient):
    """Load twitter config from database"""
    bot.tw_routingtable = build_channel_subs(
        txn,
        "iembot_twitter_oauth",
    )

    twusers = {}
    txn.execute(
        """
    SELECT iembot_account_id, access_token, access_token_secret, screen_name,
    iem_owned from
    iembot_twitter_oauth WHERE (iem_owned or (access_token is not null and
    access_token_secret is not null)) and user_id is not null and
    screen_name is not null and not disabled
    """
    )
    for row in txn.fetchall():
        user_id = row["iembot_account_id"]
        twusers[user_id] = {
            "screen_name": row["screen_name"],
            "access_token": row["access_token"],
            "access_token_secret": row["access_token_secret"],
            "iem_owned": row["iem_owned"],
        }
    bot.tw_users = twusers
    log.msg(f"load_twitter_from_db(): {txn.rowcount} oauth tokens found")


def disable_twitter_user(bot: JabberClient, user_id, errcode=0):
    """Disable the twitter subs for this user_id

    Args:
        user_id (big_id): The twitter user to disable
        errcode (int): The twitter errorcode
    """
    twuser = bot.tw_users.get(user_id)
    if twuser is None:
        log.msg(f"Failed to disable unknown twitter user_id {user_id}")
        return False
    screen_name = twuser["screen_name"]
    if twuser["iem_owned"]:
        log.msg(f"Skipping disabling of twitter for {user_id} ({screen_name})")
        return False
    bot.tw_users.pop(user_id, None)
    log.msg(
        f"Removing twitter access token for user: {user_id} ({screen_name}) "
        f"errcode: {errcode}"
    )
    df = bot.dbpool.runOperation(
        "UPDATE iembot_twitter_oauth SET updated = now(), "
        "access_token = null, access_token_secret = null "
        "WHERE user_id = %s",
        (user_id,),
    )
    df.addErrback(log.err)
    return True


def really_tweet(bot: JabberClient, user_id, twttxt, **kwargs):
    """Blocking tweet method."""
    if user_id not in bot.tw_users:
        log.msg(f"tweet() called with unknown user_id: {user_id}")
        return None
    if bot.tw_users[user_id]["access_token"] is None:
        log.msg(f"tweet() called with no access token for {user_id}")
        return None
    api = TwitterApi(
        consumer_key=bot.config["bot.twitter.consumerkey"],
        consumer_secret=bot.config["bot.twitter.consumersecret"],
        access_token_key=bot.tw_users[user_id]["access_token"],
        access_token_secret=bot.tw_users[user_id]["access_token_secret"],
    )
    # Le Sigh, api.__auth is private
    auth = OAuth1(
        bot.config["bot.twitter.consumerkey"],
        bot.config["bot.twitter.consumersecret"],
        bot.tw_users[user_id]["access_token"],
        bot.tw_users[user_id]["access_token_secret"],
    )
    oauth = OAuth1Session(
        bot.config["bot.twitter.consumerkey"],
        bot.config["bot.twitter.consumersecret"],
        bot.tw_users[user_id]["access_token"],
        bot.tw_users[user_id]["access_token_secret"],
    )
    log.msg(
        f"Tweeting {bot.tw_users[user_id]['screen_name']}({user_id}) "
        f"'{twttxt}' media:{kwargs.get('twitter_media')}"
    )
    media = kwargs.get("twitter_media")

    def _helper(params):
        """Wrap common stuff"""
        resp = api._session.post(TWEET_API, auth=auth, json=params)  # skipcq
        hh = "x-app-limit-24hour-remaining"
        log.msg(
            f"x-rate-limit-limit {resp.headers.get('x-rate-limit-limit')} + "
            f"{hh} {resp.headers.get(hh)}"
        )
        return api._ParseAndCheckTwitter(  # skipcq
            resp.content.decode("utf-8")
        )

    res = None
    try:
        params = {
            "text": twttxt,
        }
        # If we have media, we have some work to do!
        if media is not None:
            media_id = _upload_media_to_twitter(oauth, media)
            if media_id is not None:
                params["media"] = {"media_ids": [media_id]}
        res = _helper(params)
    except TwitterError as exp:
        errcode = twittererror_exp_to_code(exp)
        if errcode in [185, 187]:
            # 185: Over quota
            # 187: duplicate tweet
            return None
        if errcode in DISABLE_TWITTER_CODES:
            disable_twitter_user(bot, user_id, errcode)
            return None

        # Something bad happened with submitting this to twitter
        if str(exp).startswith("media type unrecognized"):
            # The media content hit some error, just send it without it
            log.msg(f"Sending '{kwargs.get('twitter_media')}' fail, stripping")
            params.pop("media", None)
        else:
            log.err(exp)
            # Since this called from a thread, sleeping should not jam us up
            time.sleep(10)
        res = _helper(params)
    except Exception as exp:
        log.err(exp)
        # Since this called from a thread, sleeping should not jam us up
        time.sleep(10)
        params.pop("media", None)
        res = _helper(params)
    return res


def _upload_media_to_twitter(oauth: OAuth1Session, url: str) -> str | None:
    """Upload Media to Twitter and return its ID"""
    resp = requests.get(url, timeout=30)
    if resp.status_code != 200:
        log.msg(f"Fetching `{url}` got status_code: {resp.status_code}")
        return None
    payload = resp.content
    resp = oauth.post(
        "https://api.x.com/2/media/upload",
        data={"media_category": "tweet_image"},
        files={"media": (url, payload, "image/png")},
    )
    if resp.status_code != 200:
        log.msg(f"X API got status_code: {resp.status_code} {resp.content}")
        return None
    # string required
    jresponse = resp.json()
    media_id = jresponse.get("data", {}).get("id")
    if media_id is None:
        log.msg(f"X API response did not contain id: {jresponse}")
        return None
    return str(media_id)


def twitter_errback(err, bot: JabberClient, user_id, tweettext):
    """Error callback when simple twitter workflow fails."""
    # Always log it
    log.err(err)
    errcode = twittererror_exp_to_code(err)
    if errcode in DISABLE_TWITTER_CODES:
        disable_twitter_user(bot, user_id, errcode)
    else:
        sn = bot.tw_users.get(user_id, {}).get("screen_name", "")
        msg = f"User: {user_id} ({sn})\nFailed to tweet: {tweettext}"
        email_error(err, bot, msg)


def tweet_cb(response, bot: JabberClient, twttxt, _room, myjid, user_id):
    """
    Called after success going to twitter
    """
    if response is None:
        return None
    twuser = bot.tw_users.get(user_id)
    if twuser is None:
        return response
    if "data" not in response:
        log.msg(f"Got response without data {response}")
        return None
    screen_name = twuser["screen_name"]
    url = f"https://x.com/{screen_name}/status/{response['data']['id']}"

    # Log
    df = bot.dbpool.runOperation(
        "INSERT into iembot_social_log(medium, source, resource_uri, "
        "message, response, response_code, iembot_account_id) "
        "values (%s,%s,%s,%s,%s,%s,%s)",
        ("twitter", myjid, url, twttxt, repr(response), 200, user_id),
    )
    df.addErrback(log.err)
    return response


def twittererror_exp_to_code(exp) -> int:
    """Convert a TwitterError Exception into a code.

    Args:
      exp (TwitterError): The exception to convert
    """
    errcode = None
    errmsg = str(exp)
    # brittle
    errmsg = errmsg[errmsg.find("[{") : errmsg.find("}]") + 2].replace(
        "'", '"'
    )
    try:
        errobj = json.loads(errmsg)
        errcode = errobj[0].get("code", 0)
    except Exception as exp2:
        log.msg(f"Failed to parse code TwitterError: {exp2}")
    return errcode


def tweet(bot: JabberClient, user_id, twttxt, **kwargs):
    """
    Tweet a message
    """
    twttxt = safe_twitter_text(twttxt)

    df = threads.deferToThread(
        really_tweet,
        bot,
        user_id,
        twttxt,
        **kwargs,
    )
    df.addCallback(tweet_cb, bot, twttxt, "", "", user_id)
    df.addErrback(
        twitter_errback,
        bot,
        user_id,
        twttxt,
    )
    df.addErrback(
        email_error,
        bot,
        f"User: {user_id}, Text: {twttxt} Hit double exception",
    )
    return df


def route(bot: JabberClient, channels: list, elem: Element):
    """Do the twitter work."""
    alertedPages = []
    for channel in channels:
        for user_id in bot.tw_routingtable.get(channel, []):
            if user_id not in bot.tw_users:
                log.msg(f"Failed to tweet due to no access_tokens {user_id}")
                continue
            # Require the x.twitter attribute to be set to prevent
            # confusion with some ingestors still sending tweets themself
            if not elem.x.hasAttribute("twitter"):
                continue
            if user_id in alertedPages:
                continue
            alertedPages.append(user_id)
            lat = long = None
            if (
                elem.x
                and elem.x.hasAttribute("lat")
                and elem.x.hasAttribute("long")
            ):
                lat = elem.x["lat"]
                long = elem.x["long"]
            # Finally, actually tweet, this is in basicbot
            really_tweet(
                bot,
                user_id,
                elem.x["twitter"],
                twitter_media=elem.x.getAttribute("twitter_media"),
                latitude=lat,
                longitude=long,
            )



"""Exercises Slack related tests."""

from unittest import mock

import pytest
from twisted.internet import defer
from twisted.web.server import NOT_DONE_YET
from twisted.web.test.requesthelper import DummyRequest

from iembot.slack import (
    SlackInstallChannel,
    SlackSubscribeChannel,
    load_slack_from_db,
    requests,
    send_to_slack,
)
from iembot.types import JabberClient


@defer.inlineCallbacks
def test_subscribe_channel_render(bot: JabberClient):
    """Test that we can run the render workflow."""
    ss = SlackSubscribeChannel(bot)
    request = DummyRequest([])
    request.args = {
        b"team_id": [b"T12345"],
        b"channel_id": [b"C67890"],
        b"text": [b"AFDDMX"],
    }
    bot.dbpool.runInteraction = mock.Mock(return_value=defer.succeed(None))
    result = ss.render(request)
    assert result == NOT_DONE_YET
    yield defer.succeed(None)
    assert b"Subscribed" in b"".join(request.written)


def test_install_channel_render(bot: JabberClient):
    """See if we can render the install channel."""
    ss = SlackInstallChannel(bot)
    request = DummyRequest([b""])
    result = ss.render(request)
    assert result is not None
    assert request.responseCode == 302


@pytest.mark.parametrize("database", ["iembot"])
def test_subscribe_channel(dbcursor, bot):
    """See if we can subscribe?"""
    ss = SlackSubscribeChannel(bot)
    ss.store_slack_subscription(
        dbcursor,
        "TSS",
        "CSD",
        "AFDDMX",
    )


@pytest.mark.parametrize("database", ["iembot"])
def test_gh136_slack_first_channel_subscription(dbcursor, bot):
    """See if subscribe works when we need to create a slack channel."""
    ss = SlackSubscribeChannel(bot)
    ss.store_slack_subscription(
        dbcursor,
        "TSS",  # needs to exist in iembot_slack_teams
        "CSD2",  # new channel on slack unknown to the bot so far
        "AFDDMX",
    )


@pytest.mark.parametrize("database", ["iembot"])
def test_load_from_db(dbcursor, bot: JabberClient):
    """Exercise."""
    load_slack_from_db(dbcursor, bot)


def test_send_to_slack(monkeypatch):
    """Test sending to Slack."""
    posted = {}

    def mock_post(url, headers, data):
        """Mock requests.post"""
        posted["url"] = url
        posted["headers"] = headers
        posted["data"] = data

        class MockResponse:
            """Mock response object."""

            def __init__(self):
                self.content = b'{"ok":true}'

            def raise_for_status(self):
                """Mock raise for status."""

        return MockResponse()

    monkeypatch.setattr(requests, "post", mock_post)

    bot = mock.Mock()
    bot.slack_teams = {"T12345": "xoxb-fake-token"}
    channel_id = "C67890"
    elem = mock.Mock()
    elem.x = {"twitter": "Hello, Slack!"}

    send_to_slack("xoxb-fake-token", channel_id, elem)

    assert posted["url"] == "https://slack.com/api/chat.postMessage"
    assert posted["headers"]["Authorization"] == "Bearer xoxb-fake-token"
    assert posted["headers"]["Content-Type"] == "application/json"
    import json as jsonlib

    data = jsonlib.loads(posted["data"])
    assert data["text"] == "Hello, Slack!"
    assert data["channel"] == channel_id
    assert data["mrkdwn"] is False


def test_install(bot):
    """Test the status renderer."""
    ss = SlackInstallChannel(bot)
    request = mock.Mock()
    request.setResponseCode = mock.Mock()
    request.write = mock.Mock()
    assert ss.render(request) is not None



"""Tests for iembot __init__.py."""

import iembot


def test_version():
    """Test that version is set."""
    assert iembot.__version__ is not None
    assert isinstance(iembot.__version__, str)
    # Should be either 'dev' or a version string with -dev suffix
    assert "dev" in iembot.__version__ or iembot.__version__[0].isdigit()



"""XMPP Stuff."""

from twisted.words.xish.domish import Element

from iembot.types import JabberClient


def route(bot: JabberClient, channels: list, elem: Element):
    """Do XMPP stuff."""
    alertedRooms = []
    for channel in channels:
        for room in bot.routingtable.get(channel, []):
            if room in alertedRooms:
                continue
            alertedRooms.append(room)
            elem["to"] = f"{room}@{bot.config['bot.mucservice']}"
            bot.send_groupchat_elem(elem)



"""Remove a twitter user's oauth tokens and reload iembot"""

import sys

import requests
from pyiem.database import get_dbconnc


def main(argv):
    """Run for a given username"""
    screen_name = argv[1]
    pgconn, cursor = get_dbconnc("openfire")
    cursor.execute(
        """
    DELETE from iembot_twitter_oauth where screen_name = %s
    """,
        (screen_name,),
    )
    print(
        f"Removed {cursor.rowcount} entries from the database "
        f"for screen name '{screen_name}'"
    )
    cursor.close()
    pgconn.commit()

    uri = "http://iembot:9003/reload"
    req = requests.get(uri, timeout=30)
    print(f"reloading iembot {req.text}")


if __name__ == "__main__":
    main(sys.argv)



"""Entry point for the iembot CLI.

This module replaces the historical `iembot.tac` twistd application file
with a Click-driven CLI while keeping the same Twisted service wiring.
"""

from __future__ import annotations

import json
import os
import sys

import click
from psycopg.rows import dict_row
from twisted.application import internet, service
from twisted.enterprise import adbapi
from twisted.internet import reactor
from twisted.python import log
from twisted.python.logfile import DailyLogFile
from twisted.web import server

from iembot import webservices
from iembot.bot import JabberClient
from iembot.memcache import build_memcache_client
from iembot.msghandlers import register_handler


def _load_config(path: str) -> dict:
    with open(path, encoding="utf-8") as fh:
        return json.load(fh)


def _build_dbpool(config: dict) -> adbapi.ConnectionPool:
    # Password should be set via .pgpass or environment.
    return adbapi.ConnectionPool(
        "psycopg",
        cp_reconnect=True,
        dbname=config.get("bot.dbname", "iembot"),
        host=config.get("bot.dbhost", "localhost"),
        user=config.get("bot.dbuser", "iembot"),
        gssencmode="disable",
        row_factory=dict_row,
    )


def _build_memcache_client(memcache: str):
    return build_memcache_client(memcache)


def _start_logging(logfile: str | None) -> None:
    if logfile in (None, "", "-"):
        log.startLogging(sys.stdout)
        return
    logdir = os.path.dirname(logfile) or "."
    os.makedirs(logdir, exist_ok=True)
    logname = os.path.basename(logfile)
    log.startLogging(DailyLogFile(logname, logdir))


def _write_pidfile(pidfile: str | None) -> None:
    if not pidfile:
        return
    with open(pidfile, "w", encoding="utf-8") as fh:
        fh.write(f"{os.getpid()}\n")


def _remove_pidfile(pidfile: str) -> None:
    try:
        os.remove(pidfile)
    except FileNotFoundError:
        return


def _fatal():
    # If initial bootstrap fails, stop cleanly.
    log.msg("FATAL: Stopping reactor...")
    try:
        reactor.stop()
    except Exception:
        pass


@click.group()
def main() -> None:
    """IEMBot command line interface."""


@main.command()
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True, dir_okay=False),
    default="settings.json",
    show_default=True,
    help="Path to settings.json",
)
@click.option(
    "--json-port",
    type=int,
    default=9003,
    show_default=True,
    help="Port for JSON channel requests",
)
@click.option(
    "--rss-port",
    type=int,
    default=9004,
    show_default=True,
    help="Port for RSS feed requests",
)
@click.option(
    "--memcache",
    type=str,
    default="tcp:iem-memcached:11211",
    show_default=True,
    help="Memcache server URI",
)
@click.option(
    "--maxthreads",
    type=int,
    default=512,
    show_default=True,
    help="Max Twisted threadpool size",
)
@click.option(
    "--logfile",
    "-l",
    type=str,
    default="logs/iembot.log",
    show_default=True,
    help="Twisted log file path (use '-' for stdout)",
)
@click.option(
    "--pidfile",
    "-p",
    type=str,
    default="iembot.pid",
    show_default=True,
    help="PID file path (empty to disable)",
)
@click.option(
    "--disable-slack",
    is_flag=True,
    default=False,
    help="Disable Slack message handler",
)
@click.option(
    "--disable-twitter",
    is_flag=True,
    default=False,
    help="Disable Twitter/X message handler",
)
@click.option(
    "--disable-atmosphere",
    is_flag=True,
    default=False,
    help="Disable ATmosphere(Bluesky) message handler",
)
@click.option(
    "--disable-mastodon",
    is_flag=True,
    default=False,
    help="Disable Mastodon message handler",
)
def run(
    config: str,
    json_port: int,
    rss_port: int,
    memcache: str,
    maxthreads: int,
    logfile: str,
    pidfile: str,
    disable_slack: bool,
    disable_twitter: bool,
    disable_atmosphere: bool,
    disable_mastodon: bool,
) -> None:
    """Run the IEMBot service (Twisted reactor)."""

    _start_logging(logfile)
    _write_pidfile(pidfile)

    handlers = [
        (not disable_slack, "iembot.slack", "route"),
        (not disable_twitter, "iembot.twitter", "route"),
        (not disable_atmosphere, "iembot.atmosphere", "route"),
        (not disable_mastodon, "iembot.mastodon", "route"),
        (True, "iembot.xmpp", "route"),  # Required at the moment
    ]
    for enabled, module, attr in handlers:
        if enabled:
            log.msg(f"Enabling handler: {module}.{attr}")
            mod = __import__(module, fromlist=[attr])
            register_handler(getattr(mod, attr))

    # Keep the classic Twisted Application pattern from iembot.tac.
    application = service.Application("Public IEMBOT")
    service_collection = service.IServiceCollection(application)

    settings = _load_config(config)
    dbpool = _build_dbpool(settings)
    memcache_client = _build_memcache_client(memcache)

    jabber = JabberClient("iembot", dbpool, settings, memcache_client)

    # Lame means to ensure the database is reachable before starting.
    d = dbpool.runQuery("select 1")
    d.addCallback(jabber.fire_client, service_collection)
    d.addErrback(lambda failure: (log.err(failure), _fatal()))

    # Web services (same as tac: TCPServer + setServiceParent)
    json_site = server.Site(
        webservices.JSONRootResource(jabber),
        logPath="/dev/null",
    )
    json_service = internet.TCPServer(json_port, json_site)
    json_service.setServiceParent(service_collection)

    rss_site = server.Site(
        webservices.RSSRootResource(jabber),
        logPath="/dev/null",
    )
    rss_service = internet.TCPServer(rss_port, rss_site)
    rss_service.setServiceParent(service_collection)

    # Threadpool tuning (kept from tac)
    reactor.getThreadPool().adjustPoolsize(maxthreads=maxthreads)

    # Ensure services stop on shutdown.
    reactor.addSystemEventTrigger(
        "before",
        "shutdown",
        service_collection.stopService,
    )
    reactor.addSystemEventTrigger("before", "shutdown", dbpool.close)
    if pidfile:
        reactor.addSystemEventTrigger(
            "before",
            "shutdown",
            _remove_pidfile,
            pidfile,
        )
    if hasattr(memcache_client, "disconnect"):
        reactor.addSystemEventTrigger(
            "before",
            "shutdown",
            memcache_client.disconnect,
        )

    # Start services and run the reactor.
    service_collection.startService()
    if not reactor.running:
        reactor.run()


if __name__ == "__main__":
    main()



"""Placeholder."""

import os
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from typing import NamedTuple

try:
    __version__ = version("iembot")
    pkgdir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    if not pkgdir.endswith("site-packages"):
        __version__ += "-dev"
except PackageNotFoundError:
    # package is not installed
    __version__ = "dev"

DATADIR = Path(__file__).parent / "data"


class ROOM_LOG_ENTRY(NamedTuple):
    seqnum: int
    timestamp: str
    log: str
    author: str
    product_id: str
    product_text: str
    txtlog: str



"""Centralized Testing Stuff."""

import random
from collections import defaultdict
from unittest import mock

import pytest
from pyiem.database import get_dbconnc

from iembot.bot import JabberClient


@pytest.fixture
def bot():
    """A bot."""
    iembot = JabberClient(f"iembot_{random.randint(0, 1000000)}", mock.Mock())
    iembot.config = defaultdict(str)
    iembot.xmlstream = mock.Mock()
    return iembot


@pytest.fixture
def dbcursor(database):
    """Yield a cursor for the given database."""
    dbconn, cursor = get_dbconnc(database, user="mesonet")
    yield cursor
    dbconn.close()



"""Tests for src/iembot/main.py"""

import json
import os
import sys
import types
from unittest import mock

from click.testing import CliRunner

import iembot.main as main_mod


def test_load_settings_reads_json(tmp_path):
    d = {"foo": "bar"}
    f = tmp_path / "settings.json"
    f.write_text(json.dumps(d))
    assert main_mod._load_config(str(f)) == d


def test_write_and_remove_pidfile(tmp_path):
    pidfile = tmp_path / "pid"
    main_mod._write_pidfile(str(pidfile))
    assert pidfile.read_text().strip() == str(os.getpid())
    main_mod._remove_pidfile(str(pidfile))
    assert not pidfile.exists()
    # Should not raise if already gone
    main_mod._remove_pidfile(str(pidfile))


def test_start_logging_stdout(monkeypatch):
    called = {}
    monkeypatch.setattr(
        main_mod.log,
        "startLogging",
        lambda arg: called.setdefault("stdout", arg),
    )
    main_mod._start_logging("-")
    assert called["stdout"] is sys.stdout


def test_start_logging_file(monkeypatch, tmp_path):
    called = {}

    class DummyDLF:
        def __init__(self, name, _dir):
            called["name"] = name
            called["dir"] = _dir

    monkeypatch.setattr(main_mod, "DailyLogFile", DummyDLF)
    monkeypatch.setattr(
        main_mod.log,
        "startLogging",
        lambda arg: called.setdefault("file", arg),
    )
    logfile = tmp_path / "foo" / "bar.log"
    main_mod._start_logging(str(logfile))
    assert called["name"] == "bar.log"
    assert os.path.basename(str(logfile)) == "bar.log"
    assert called["dir"] == str(tmp_path / "foo")


def test_build_dbpool(monkeypatch):
    fakepool = object()

    def fakepool_ctor(*_a, **_k):
        return fakepool

    monkeypatch.setattr(
        main_mod, "adbapi", types.SimpleNamespace(ConnectionPool=fakepool_ctor)
    )
    settings = {
        "databaserw": {
            "openfire": "db",
            "host": "h",
            "password": "p",
            "user": "u",
        }
    }
    assert main_mod._build_dbpool(settings) is fakepool


def test_build_memcache_client():
    result = main_mod._build_memcache_client("tcp:foo:1234")
    assert result


def test_fatal_stops_reactor(monkeypatch):
    called = {}
    monkeypatch.setattr(
        main_mod,
        "reactor",
        types.SimpleNamespace(stop=lambda: called.setdefault("stopped", True)),
    )
    main_mod._fatal()
    assert called["stopped"]


def taest_cli_help():
    runner = CliRunner()
    result = runner.invoke(main_mod.main, ["--help"])
    assert result.exit_code == 0
    assert "IEMBot command line interface" in result.output


def taest_cli_run_help():
    runner = CliRunner()
    result = runner.invoke(main_mod.main, ["run", "--help"])
    assert result.exit_code == 0
    assert "Run the IEMBot service" in result.output


def test_cli_run(tmp_path):
    runner = CliRunner()
    with mock.patch("iembot.main._start_logging"):
        runner.invoke(
            main_mod.main,
            ["run", "--logfile", tmp_path / "bah.log"],
            catch_exceptions=False,
        )



"""Test iembot.xmpp"""

from twisted.words.xish.domish import Element

from iembot.xmpp import route


def test_route(bot):
    """Can we route stuff, yes we can."""
    msg = Element((None, "message"))
    msg["from"] = "test@example.com"
    msg["to"] = "dmxchat@localhost"
    msg.addElement("body", content="Hello, world!")
    route(bot, ["DMX"], msg)



"""Test iembot.mastodon"""

from unittest import mock

import mastodon
import pytest

from iembot.bot import JabberClient
from iembot.mastodon import (
    disable_mastodon_user,
    load_mastodon_from_db,
    mastodon_errback,
    toot,
    toot_cb,
)


def test_toot_cb_no_response():
    """Test toot_cb with None response."""
    bot = mock.Mock()
    result = toot_cb(None, bot, "text", "123")
    assert result is None


def test_toot_cb_no_user():
    """Test toot_cb with unknown user."""
    bot = mock.Mock()
    bot.md_users = {}
    result = toot_cb({"content": "test"}, bot, "text", "999")
    assert result == {"content": "test"}


def test_toot_cb_no_content():
    """Test toot_cb with response missing content."""
    bot = mock.Mock()
    bot.md_users = {"123": {"screen_name": "test"}}
    result = toot_cb({"error": "bad"}, bot, "text", "123")
    assert result is None


def test_toot_cb_success():
    """Test toot_cb with successful response."""
    bot = mock.Mock()
    bot.md_users = {"123": {"screen_name": "testuser"}}
    bot.dbpool.runOperation.return_value = mock.Mock()
    response = {
        "content": "test",
        "url": "https://mastodon.social/@test/123",
        "account": {},
    }
    result = toot_cb(response, bot, "text", "123")
    assert result == {
        "content": "test",
        "url": "https://mastodon.social/@test/123",
    }
    bot.dbpool.runOperation.assert_called_once()


@pytest.mark.parametrize("database", ["iembot"])
def test_load_mastodon_from_db(dbcursor):
    """Test the method."""
    bot = JabberClient(None, None)
    load_mastodon_from_db(dbcursor, bot)


def test_util_toot():
    """Test the method."""
    bot = JabberClient(None, None)
    bot.md_users = {
        "123": {
            "screen_name": "iembot",
            "access_token": "123",
            "api_base_url": "https://localhost",
        }
    }
    toot(bot, "123", "test", sleep=0)


def test_mastodon_errback_not_found():
    """Test mastodon_errback with 404 error."""
    bot = mock.Mock()
    bot.md_users = {"123": {"screen_name": "test", "iem_owned": False}}
    bot.dbpool.runOperation.return_value = mock.Mock()
    err = mastodon.errors.MastodonNotFoundError("Not found")
    mastodon_errback(err, bot, "123", "test toot")
    # User should be disabled
    assert "123" not in bot.md_users


def test_mastodon_errback_unauthorized():
    """Test mastodon_errback with 401 error."""
    bot = mock.Mock()
    bot.md_users = {"123": {"screen_name": "test", "iem_owned": False}}
    bot.dbpool.runOperation.return_value = mock.Mock()
    err = mastodon.errors.MastodonUnauthorizedError("Unauthorized")
    mastodon_errback(err, bot, "123", "test toot")
    # User should be disabled
    assert "123" not in bot.md_users


def test_disable_mastodon_user_unknown():
    """Test disable_mastodon_user with unknown user."""
    bot = mock.Mock()
    bot.md_users = {}
    assert disable_mastodon_user(bot, "unknown") is False


def test_disable_mastodon_user_iem_owned():
    """Test disable_mastodon_user with IEM owned account."""
    bot = mock.Mock()
    bot.md_users = {"123": {"screen_name": "iembot", "iem_owned": True}}
    assert disable_mastodon_user(bot, "123") is False


def test_disable_mastodon_user_success():
    """Test disable_mastodon_user with valid user."""
    bot = mock.Mock()
    bot.md_users = {"123": {"screen_name": "testuser", "iem_owned": False}}
    bot.dbpool.runOperation.return_value = mock.Mock()
    result = disable_mastodon_user(bot, "123", errcode=401)
    assert result is True
    assert "123" not in bot.md_users



"""Tests, gasp"""

import os
from unittest import mock

import pytest
from twisted.words.xish.domish import Element

import iembot.util as botutil
from iembot.bot import JabberClient
from iembot.msghandlers import process_groupchat


def test_remove_control_characters():
    """Test remove_control_characters function."""
    # Test control characters are removed
    assert botutil.remove_control_characters("hello\x00world") == "helloworld"
    assert botutil.remove_control_characters("test\x07data") == "testdata"
    assert botutil.remove_control_characters("clean text") == "clean text"
    # Tab and newline should be preserved (not in the removed range)
    assert botutil.remove_control_characters("tab\there") == "tab\there"


def test_purge_logs(tmp_path):
    """Test purge_logs function."""
    bot = mock.Mock()
    bot.config = {"bot.purge_xmllog_days": "1"}
    # Create a temporary log directory
    log_dir = tmp_path / "logs"
    log_dir.mkdir()
    # Create an old log file that should be deleted
    old_log = log_dir / "xmllog.2020_01_01"
    old_log.touch()
    # Create a recent log file that should be kept
    recent_log = log_dir / "xmllog.2099_01_01"
    recent_log.touch()

    # Change to temp directory so glob works
    orig_cwd = os.getcwd()
    try:
        os.chdir(tmp_path)
        botutil.purge_logs(bot)
        assert not old_log.exists()
        assert recent_log.exists()
    finally:
        os.chdir(orig_cwd)


def test_channels_room_list():
    """Test channels_room_list."""
    bot = mock.Mock()
    bot.routingtable = {
        "ABC": ["room1", "room2"],
        "DEF": ["room1"],
        "GHI": ["room3"],
    }
    botutil.channels_room_list(bot, "room1")
    bot.send_groupchat.assert_called_once()
    call_args = bot.send_groupchat.call_args
    assert "room1" in call_args[0][0]
    assert "2 channels" in call_args[0][1]


def test_load_chatlog(tmp_path, bot: JabberClient):
    """Test our pickling fun."""
    print(bot.seqnum)
    bot.picklefile = tmp_path / "chatlog.pkl"
    bot.save_chatlog()
    botutil.load_chatlog(bot)
    assert bot.seqnum == 0

    # Create a faked message
    message = Element(("jabber:client", "message"))
    message["from"] = f"lotchat@{bot.conference}/iembot"
    message["type"] = "groupchat"
    message.addElement("body", None, "Hello World")
    xelem = message.addElement("x", "nwschat:nwsbot")
    xelem["channels"] = "ABC"
    process_groupchat(bot, message)
    bot.save_chatlog()
    botutil.load_chatlog(bot)
    assert bot.seqnum == 1


@pytest.mark.parametrize("database", ["iembot"])
def test_load_chatrooms_fromdb(dbcursor):
    """Can we load up chatroom details?"""
    bot = mock.Mock()
    bot.rooms = {}
    botutil.load_chatrooms_from_db(dbcursor, bot, True)
    assert bot


def test_daily_timestamp(bot: JabberClient):
    """Does the daily timestamp algo return a deferred."""
    assert botutil.daily_timestamp(bot) is not None


def test_htmlentites():
    """Do replacements work?"""
    assert botutil.htmlentities("<") == "&lt;"
    assert botutil.htmlentities("<>") == "&lt;&gt;"



"""Tests for iembot atmosphere module."""

from unittest import mock

from iembot.atmosphere import ATManager, at_send_message, route
from iembot.types import JabberClient


def test_at_send_message_unknown_user(bot: JabberClient):
    """Test at_send_message with unknown user."""
    # Should not raise an error
    route(bot, "unknown_user", "test message")


def test_at_send_message_no_handle(bot: JabberClient):
    """Test at_send_message with user that has no at_handle."""
    bot.at_users = {"123": {"at_handle": None}}
    # Should not raise an error
    route(bot, "123", "test message")


def test_at_send_message_with_handle(bot: JabberClient):
    """Test at_send_message with valid at_handle."""
    bot.at_users = {"123": {"at_handle": "test.bsky.social"}}
    bot.at_manager = ATManager()
    bot.at_manager.at_clients["test.bsky.social"] = mock.Mock()
    bot.at_manager.submit = mock.Mock()
    at_send_message(bot, "123", "test message", extra_key="value")
    bot.at_manager.submit.assert_called_once_with(
        "test.bsky.social", {"msg": "test message", "extra_key": "value"}
    )


def test_atmanager_add_client():
    """Test ATManager add_client."""
    manager = ATManager()
    with mock.patch("iembot.atmosphere.ATWorkerThread") as mock_thread:
        mock_instance = mock.Mock()
        mock_thread.return_value = mock_instance
        manager.add_client("test.bsky.social", "password123")
        mock_thread.assert_called_once()
        mock_instance.start.assert_called_once()
        assert "test.bsky.social" in manager.at_clients


def test_atmanager_add_client_duplicate():
    """Test ATManager doesn't add duplicate clients."""
    manager = ATManager()
    manager.at_clients["test.bsky.social"] = mock.Mock()
    with mock.patch("iembot.atmosphere.ATWorkerThread") as mock_thread:
        manager.add_client("test.bsky.social", "password123")
        mock_thread.assert_not_called()


def test_atmanager_submit():
    """Test ATManager submit."""
    manager = ATManager()
    mock_queue = mock.Mock()
    mock_thread = mock.Mock()
    mock_thread.queue = mock_queue
    manager.at_clients["test.bsky.social"] = mock_thread

    manager.submit("test.bsky.social", {"msg": "Hello"})
    mock_queue.put.assert_called_once_with({"msg": "Hello"})



"""Type definitions for iembot."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol

if TYPE_CHECKING:
    from datetime import datetime


class JabberClient(Protocol):
    """Structural type for JabberClient to avoid import cycles."""

    startup_time: datetime
    picklefile: str
    name: str
    dbpool: Any
    memcache_client: Any | None
    config: dict[str, Any]
    outstanding_pings: list
    rooms: dict[str, dict[str, Any]]
    chatlog: dict[str, Any]
    seqnum: int
    routingtable: dict[str, list[str]]
    at_manager: Any
    at_users: dict[str, dict[str, Any]]
    at_routingtable: dict[str, list[str]]
    tw_users: dict[str, dict[str, Any]]
    tw_routingtable: dict[str, list[str]]
    md_users: dict[str, dict[str, Any]]
    md_routingtable: dict[str, list[str]]
    slack_teams: dict[str, str]
    slack_routingtable: dict[str, list[str]]
    webhooks_routingtable: dict[str, list[str]]
    xmlstream: Any | None
    firstlogin: bool
    xmllog: Any
    myjid: Any | None
    ingestjid: Any | None
    conference: str | None
    email_timestamps: list[datetime]
    keepalive_lc: Any | None
    fortunes: list[str]

    def fire_client(self, xs: Any, service_collection: Any) -> None:
        """Fire up the client."""



"""Try to test the webservices."""

from iembot import webservices


def test_status(bot):
    """Test the status renderer."""
    ss = webservices.StatusChannel(bot)
    assert ss.render(None) is not None


def test_api(bot):
    """Can we import API?"""
    res = webservices.wfo_rss(bot, "dmxchat")
    assert res is not None



"""Slack integration work."""

import json
import urllib

import requests
from twisted.internet import threads
from twisted.python import log
from twisted.web import resource
from twisted.web.server import NOT_DONE_YET
from twisted.words.xish.domish import Element

from iembot.types import JabberClient
from iembot.util import build_channel_subs


def send_to_slack(access_token: str, channel_id: str, elem: Element):
    """Send a message to Slack, called from thread."""
    payload = {
        "text": elem.x["twitter"],
        "mrkdwn": False,
        "channel": channel_id,
        "unfurl_links": False,
        "unfurl_media": False,
    }
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {access_token}",
    }
    log.msg("Posting to slack")
    resp = requests.post(
        "https://slack.com/api/chat.postMessage",
        headers=headers,
        data=json.dumps(payload),
    )
    log.msg(f"Got response {resp} {resp.content}")
    resp.raise_for_status()


def load_slack_from_db(txn, bot: JabberClient):
    """Load the Slack integration."""
    bot.slack_routingtable = build_channel_subs(
        txn,
        "iembot_slack_team_channels",
    )

    txn.execute(
        """
    select iembot_account_id, c.channel_id, t.access_token from
    iembot_slack_teams t JOIN iembot_slack_team_channels c on
    (t.team_id = c.team_id)
        """
    )
    teams = {}
    xref = {}
    for row in txn.fetchall():
        teams[row["iembot_account_id"]] = {
            "access_token": row["access_token"],
            "channel_id": row["channel_id"],
        }
        xref[row["iembot_account_id"]] = row["channel_id"]

    bot.slack_teams = teams
    log.msg(f"Loaded {len(teams)} Slack teams")


def route(bot: JabberClient, channels: list, elem: Element):
    """Do Slack message routing."""
    alerted = []
    for channel in channels:
        for iembot_account_id in bot.slack_routingtable.get(channel, []):
            if iembot_account_id in alerted:
                continue
            alerted.append(iembot_account_id)
            log.msg("Attempting slack send...")
            meta = bot.slack_teams[iembot_account_id]
            d = threads.deferToThread(
                send_to_slack, meta["access_token"], meta["channel_id"], elem
            )
            d.addErrback(log.msg)


class SlackSubscribeChannel(resource.Resource):
    """respond to /subscribe requests"""

    def __init__(self, iembot: JabberClient):
        """Constructor"""
        resource.Resource.__init__(self)
        self.iembot = iembot

    def store_slack_subscription(
        self,
        txn,
        team_id: str,
        channel_id: str,
        subkey: str,
    ):
        """write to database."""
        log.msg(
            f"Handing slack subscription for T:`{team_id}` TC:`{channel_id}` "
            f"C:`{subkey}`"
        )
        # Step 1, ensure that the slack channel exists
        txn.execute(
            """
    select iembot_account_id from iembot_slack_team_channels
    where team_id = %s and channel_id = %s
            """,
            (team_id, channel_id),
        )
        res = txn.fetchall()
        if not res:
            log.msg(
                f"Slack Team: {team_id} Channel: {channel_id} does not exist "
                "creating..."
            )
            txn.execute(
                """
                insert into iembot_slack_team_channels
                (iembot_account_id, team_id, channel_id)
                values (
                    (select create_iembot_account('slack')), %s, %s)
                returning iembot_account_id
                """,
                (team_id, channel_id),
            )
            res = txn.fetchall()
        iembot_account_id = res[0]["iembot_account_id"]
        txn.execute(
            """
            insert into iembot_subscriptions(iembot_account_id, channel_id)
            values (
                %s,
                (select get_or_create_iembot_channel_id(%s))
            )
            on conflict do nothing
            """,
            (iembot_account_id, subkey),
        )

    def render(self, request):
        """Answer the call."""
        team_id = request.args.get(b"team_id", [b""])[0].decode("ascii")
        channel_id = request.args.get(b"channel_id", [b""])[0].decode("ascii")
        subkey = request.args.get(b"text", [b""])[0].decode("ascii")
        defer = self.iembot.dbpool.runInteraction(
            self.store_slack_subscription,
            team_id,
            channel_id,
            subkey,
        )
        defer.addErrback(
            lambda _: request.write(b"Error processing subscription")
        )
        defer.addCallback(
            lambda _: request.write(f"Subscribed to {subkey}".encode("ascii"))
        )
        defer.addBoth(lambda _: request.finish())
        defer.addCallback(lambda _: self.iembot.load_slack())

        return NOT_DONE_YET


class SlackUnsubscribeChannel(resource.Resource):
    """respond to /unsubscribe requests"""

    def __init__(self, iembot: JabberClient):
        """Constructor"""
        resource.Resource.__init__(self)
        self.iembot = iembot

    def remove_slack_subscription(
        self,
        txn,
        team_id,
        channel_id,
        subkey,
    ):
        """write to database."""
        txn.execute(
            """
            delete from iembot_subscriptions
            where iembot_account_id = (
            select iembot_account_id from iembot_slack_team_channels where
            team_id = %s and channel_id = %s) and
            channel_id = get_or_create_iembot_channel_id(%s)
            """,
            (team_id, channel_id, subkey),
        )

    def render(self, request):
        """Answer the call."""
        team_id = request.args.get(b"team_id", [b""])[0].decode("ascii")
        channel_id = request.args.get(b"channel_id", [b""])[0].decode("ascii")
        subkey = request.args.get(b"text", [b""])[0].decode("ascii")
        defer = self.iembot.dbpool.runInteraction(
            self.remove_slack_subscription,
            team_id,
            channel_id,
            subkey,
        )
        defer.addCallback(
            lambda _: request.write(
                f"Unsubscribed from {subkey}".encode("ascii")
            )
        )
        defer.addCallback(lambda _: request.finish())
        defer.addCallback(lambda _: self.iembot.load_slack())

        return NOT_DONE_YET


class SlackOauthChannel(resource.Resource):
    """respond to /oauth requests"""

    def __init__(self, iembot: JabberClient):
        """Constructor"""
        resource.Resource.__init__(self)
        self.iembot = iembot

    def store_slack_team(self, txn, data):
        """Save."""

    def _eb_oauth(self, failure, request):
        """Errback on oauth call failure."""
        log.msg(f"OAuth error: {failure}")
        request.setResponseCode(500)
        request.write(b"OAuth error")
        request.finish()

    def _cb_oauth(self, message, request):
        """Errback on oauth call success."""
        log.msg(f"OAuth success: {message}")
        request.setResponseCode(200)
        request.write(b"OAuth success")
        request.finish()

    def do_request_in_thread(self, txn, data):
        """Run a request from a dbthread, sigh."""
        token_url = "https://slack.com/api/oauth.v2.access"
        resp = requests.post(
            token_url,
            data,
        )
        jdata = resp.json()
        if jdata.get("ok"):
            txn.execute(
                """
        insert into iembot_slack_teams
            (team_id, team_name, access_token, bot_user_id)
        values (%s, %s, %s, %s)
        on conflict (team_id) do update set
        access_token = excluded.access_token,
        bot_user_id = excluded.bot_user_id,
        team_name = excluded.team_name
        """,
                (
                    jdata["team"]["id"],
                    jdata["team"]["name"],
                    jdata["access_token"],
                    jdata["bot_user_id"],
                ),
            )
        return jdata

    def render(self, request):
        """Answer the call."""
        # Get the posted code parameter
        code = request.args.get(b"code")[0]
        if not code:
            request.setResponseCode(400)
            return b"Missing code parameter"
        # Exchange code for token
        data = {
            "client_id": self.iembot.config["bot.slack.client_id"],
            "client_secret": self.iembot.config["bot.slack.client_secret"],
            "code": code.decode("ascii"),
            "redirect_uri": self.iembot.config["bot.slack.redirect_uri"],
        }
        defer = self.iembot.dbpool.runInteraction(
            self.do_request_in_thread, data
        )
        defer.addCallback(self._cb_oauth, request)
        defer.addCallback(lambda _: self.iembot.load_slack())
        defer.addErrback(self._eb_oauth, request)
        return NOT_DONE_YET


class SlackInstallChannel(resource.Resource):
    """respond to /oauth requests"""

    def __init__(self, iembot: JabberClient):
        """Constructor"""
        resource.Resource.__init__(self)
        self.iembot = iembot

    def render(self, request):
        """Answer the call."""
        params = {
            "client_id": self.iembot.config["bot.slack.client_id"],
            "scope": (
                "chat:write,channels:read,groups:read,"
                "im:read,mpim:read,commands"
            ),
            "user_scope": "",
            "redirect_uri": self.iembot.config["bot.slack.redirect_uri"],
        }
        url = (
            "https://slack.com/oauth/v2/authorize?"
            f"{urllib.parse.urlencode(params)}"
        )
        # Send 302 header
        request.setResponseCode(302)
        request.setHeader("Location", url)
        return b"Redirecting to Slack"



"""The ATmosphere Worker.

Need to do some hackery to workaround threadsafety issues with the
atproto module and my general lack of understanding of how to do
threadsafety in Python.  So this is on me.

"""

import threading
from queue import Queue

import httpx
from atproto import Client
from atproto_client.utils import TextBuilder
from twisted.python import log
from twisted.words.xish.domish import Element

from iembot.types import JabberClient
from iembot.util import build_channel_subs


class ATWorkerThread(threading.Thread):
    """The Worker."""

    def __init__(self, queue: Queue, at_handle: str, at_password: str):
        """Constructor."""
        threading.Thread.__init__(self)
        self.queue = queue
        self.at_handle = at_handle
        self.at_password = at_password
        self.logged_in = False
        self.client = Client()
        self.daemon = True  # Don't block on shutdown

    def run(self):
        """Listen for messages from the queue."""
        while True:
            # Grab the message from the queue
            message = self.queue.get()
            if message is None:
                break
            # Process the message
            try:
                self.process_message(message)
            except Exception as exp:
                log.err(exp)
                self.logged_in = False
            self.queue.task_done()

    def process_message(self, msgdict: dict):
        """Process the message."""
        media = msgdict.get("twitter_media")
        img = None
        if media is not None:
            try:
                resp = httpx.get(media, timeout=30)
                resp.raise_for_status()
                img = resp.content
                # AT has a size limit of 976.56KB
                if len(img) > 1_000_000:
                    log.msg(f"{media} is too large({len(img)}) for AT")
                    img = None
            except Exception as exp:
                log.err(exp)

        # Do we need to login?
        if not self.logged_in:
            log.msg(f"Logging in as {self.at_handle}...")
            me = self.client.login(self.at_handle, self.at_password)
            log.msg(repr(me))
            self.logged_in = True

        msg = msgdict["msg"]
        if msg.find("http") > -1:
            parts = msg.split("http")
            msg = TextBuilder().text(parts[0]).link("Link", f"http{parts[1]}")

        if img:
            res = self.client.send_image(
                msg, image=img, image_alt="IEMBot Image TBD"
            )
        else:
            res = self.client.send_post(msg)
        # for now
        log.msg(repr(res))


class ATManager:
    """Ensure the creation of clients and submission of tasks is threadsafe."""

    def __init__(self):
        """Constructor."""
        self.at_clients = {}
        self.lock = threading.Lock()

    def add_client(self, at_handle: str, at_password: str):
        """Add a new client, if necessary."""
        if at_handle in self.at_clients:
            return
        with self.lock:
            self.at_clients[at_handle] = ATWorkerThread(
                Queue(), at_handle, at_password
            )
            self.at_clients[at_handle].start()

    def submit(self, at_handle: str, message: dict):
        """Submit a message to the client."""
        self.at_clients[at_handle].queue.put(message)


def load_atmosphere_from_db(txn, bot: JabberClient):
    """Query database for our config."""

    bot.at_routingtable = build_channel_subs(
        txn,
        "iembot_atmosphere_accounts",
    )

    users = {}
    txn.execute(
        """
    SELECT iembot_account_id, handle, app_pass from
    iembot_atmosphere_accounts
    """
    )
    for row in txn.fetchall():
        user_id = row["iembot_account_id"]
        users[user_id] = {
            "at_handle": row["handle"],
        }
        bot.at_manager.add_client(row["handle"], row["app_pass"])
    bot.at_users = users
    log.msg(f"load_atmosphere_from_db(): {txn.rowcount} accounts found")


def at_send_message(bot: JabberClient, iembot_account_id, msg: str, **kwargs):
    """Send a message to the ATmosphere."""
    at_handle = bot.at_users.get(iembot_account_id, {}).get("at_handle")
    if at_handle is None:
        return
    message = {"msg": msg}
    message.update(kwargs)
    bot.at_manager.submit(at_handle, message)


def route(bot: JabberClient, channels: list, elem: Element):
    """Do the message routing."""
    alerted = []
    for channel in channels:
        for iembot_account_id in bot.at_routingtable.get(channel, []):
            if iembot_account_id in alerted:
                log.msg(f"{iembot_account_id} already alerted...")
                continue
            alerted.append(iembot_account_id)
            at_send_message(bot, iembot_account_id, elem.x["twitter"])



"""test twitter."""

from unittest import mock

from twisted.python.failure import Failure
from twitter import TwitterError

from iembot.twitter import (
    disable_twitter_user,
    safe_twitter_text,
    tweet_cb,
    twitter_errback,
    twittererror_exp_to_code,
)


def test_error_conversion():
    """Test that we can convert errors."""
    err = TwitterError("BLAH")
    assert twittererror_exp_to_code(err) is None
    err = TwitterError(
        "[{'code': 185, 'message': 'User is over daily status update limit.'}]"
    )
    assert twittererror_exp_to_code(err) == 185
    assert twittererror_exp_to_code(Failure(err)) == 185


def test_disable_twitter_user_unknown():
    """Test disable_twitter_user with unknown user."""
    bot = mock.Mock()
    bot.tw_users = {}
    assert disable_twitter_user(bot, "unknown") is False


def test_disable_twitter_user_iem_owned():
    """Test disable_twitter_user with IEM owned account."""
    bot = mock.Mock()
    bot.tw_users = {"123": {"screen_name": "iembot", "iem_owned": True}}
    assert disable_twitter_user(bot, "123") is False


def test_disable_twitter_user_success():
    """Test disable_twitter_user with valid user."""
    bot = mock.Mock()
    bot.tw_users = {"123": {"screen_name": "testuser", "iem_owned": False}}
    bot.dbpool.runOperation.return_value = mock.Mock()
    result = disable_twitter_user(bot, "123", errcode=185)
    assert result is True
    assert "123" not in bot.tw_users


def test_tweettext():
    """Are we doing the right thing here"""
    msgin = (
        "At 1:30 PM, 1 WNW Lake Mills [Winnebago Co, IA] TRAINED "
        "SPOTTER reports TSTM WND GST of E61 MPH. SPOTTER MEASURED "
        "61 MPH WIND GUST. HIS CAR DOOR WAS ALSO CAUGHT BY THE WIND "
        "WHEN HE WAS OPENING THE DOOR, PUSHING THE DOOR INTO HIS FACE. "
        "THIS CONTACT BR.... "
        "https://iem.local/lsr/#DMX/201807041830/201807041830"
    )
    msgout = safe_twitter_text(msgin)
    assert msgout == msgin


def test_tweet_unescape():
    """Test that we remove html entities from string."""
    msg = "Hail &gt; 2.0 INCHES"
    ans = "Hail > 2.0 INCHES"
    assert safe_twitter_text(msg) == ans


def test_tweet_cb_no_data():
    """Test tweet_cb with response missing data."""
    bot = mock.Mock()
    bot.tw_users = {"123": {"screen_name": "test"}}
    result = tweet_cb({"error": "bad"}, bot, "text", "room", "jid", "123")
    assert result is None


def test_tweet_cb_success():
    """Test tweet_cb with successful response."""
    bot = mock.Mock()
    bot.tw_users = {"123": {"screen_name": "testuser"}}
    bot.dbpool.runOperation.return_value = mock.Mock()
    response = {"data": {"id": "tweet123"}}
    result = tweet_cb(response, bot, "text", "room", "jid", "123")
    assert result == response
    bot.dbpool.runOperation.assert_called_once()


def test_tweet_cb_no_response():
    """Test tweet_cb with None response."""
    bot = mock.Mock()
    result = tweet_cb(None, bot, "text", "room", "jid", "123")
    assert result is None


def test_tweet_cb_no_user():
    """Test tweet_cb with unknown user."""
    bot = mock.Mock()
    bot.tw_users = {}
    result = tweet_cb(
        {"data": {"id": "123"}}, bot, "text", "room", "jid", "999"
    )
    assert result == {"data": {"id": "123"}}


def test_twitter_errback_disable_code():
    """Test twitter_errback with disable code."""
    bot = mock.Mock()
    bot.tw_users = {"123": {"screen_name": "test", "iem_owned": False}}
    bot.dbpool.runOperation.return_value = mock.Mock()
    err = TwitterError("[{'code': 89, 'message': 'Expired token'}]")
    twitter_errback(err, bot, "123", "test tweet")
    # User should be disabled
    assert "123" not in bot.tw_users


def test_twittertext():
    """Tricky stuff when URLs are included"""
    msg = (
        "Tropical Storm #Danny Intermediate ADVISORY 23A issued. "
        "Outer rainbands spreading across the southern leeward "
        "islands. http://go.usa.gov/W3H"
    )
    msg2 = safe_twitter_text(msg)
    assert msg2 == (
        "Tropical Storm #Danny Intermediate ADVISORY 23A "
        "issued. Outer rainbands spreading across the "
        "southern leeward islands. http://go.usa.gov/W3H"
    )


def test_safe_twitter_text_long_with_url():
    """Test safe_twitter_text with long text and URL."""
    url = "https://example.com/very/long/path"
    # Create a message that's over the limit
    msg = "A" * 300 + " " + url
    result = safe_twitter_text(msg)
    assert url in result
    assert "..." in result


def test_safe_twitter_text_sections():
    """Test safe_twitter_text with for/till pattern."""
    url = "https://example.com/alert"
    msg = (
        "Tornado Warning for Very Long County Name and Another County Name "
        "till 3:00 PM CDT " + url
    )
    result = safe_twitter_text(msg)
    assert url in result


def test_safe_twitter_text_double_spaces():
    """Test safe_twitter_text removes double spaces."""
    msg = "Hello  World   Test"
    result = safe_twitter_text(msg)
    assert result == "Hello World Test"


def test_safe_twitter_text_no_url():
    """Test safe_twitter_text with no URL and under limit."""
    msg = "Short message without URL"
    result = safe_twitter_text(msg)
    assert result == msg


def test_safe_twitter_text_url_only_counts_25():
    """Test that URL counts as 25 chars for limit calculation."""
    # 250 chars + 25 for URL = 275, under 280
    msg = "A" * 250 + " https://example.com"
    result = safe_twitter_text(msg)
    assert result == msg



"""Test bot API."""

from unittest.mock import Mock

from iembot.bot import JabberClient


def test_bot_apis(bot: JabberClient):
    """Exercise things within the bot."""
    bot.fire_client(None, Mock())
    bot.send_help_message("admin")
    bot.send_groupchat_help("dmxchat")


def test_authd_api(bot: JabberClient):
    """Call authd."""
    xs = Mock()
    bot.connected(xs)
    bot.authd()


def test_xml(bot):
    msg = bot.send_groupchat("roomname", "Hello Friend")
    assert msg is not None

    msg = bot.send_groupchat("roomname", "Hello Friend &")
    assert msg is not None

    msg = bot.send_groupchat("roomname", "Hello Friend &&amp;")
    assert msg is not None



"""Test things done with channel subs."""

import pytest

from iembot.bot import JabberClient
from iembot.util import (
    channels_room_add,
    channels_room_del,
    channels_room_list,
)


@pytest.mark.parametrize("database", ["iembot"])
def test_room_list(bot: JabberClient, dbcursor):
    """Test listing of channel subscriptions for the room."""

    def _local(room, _msg):
        """Interception."""
        assert room == "dmxchat"

    channels_room_add(dbcursor, bot, "dmxchat", "XXX")

    bot.send_groupchat = _local
    channels_room_list(bot, "dmxchat")

    channels_room_del(dbcursor, bot, "dmxchat", "XXX")



"""Memcache client helpers."""

from __future__ import annotations

from pymemcache.client.base import Client as MemcacheClient
from twisted.internet import threads


def parse_memcache_addr(memcache: str) -> tuple[str, int]:
    """."""
    memcache = memcache.removeprefix("tcp:")
    host = memcache
    port = 11211
    if ":" in memcache:
        host, port_str = memcache.rsplit(":", 1)
        try:
            port = int(port_str)
        except ValueError:
            host = memcache
    if not host:
        host = "localhost"
    return host, port


class ThreadedMemcacheClient:
    def __init__(self, host: str, port: int) -> None:
        self._client = MemcacheClient((host, port))

    def get(self, key: bytes):
        def _get():
            data = self._client.get(key)
            return (0, data)

        return threads.deferToThread(_get)

    def disconnect(self) -> None:
        try:
            self._client.close()
        except Exception:
            pass


def build_memcache_client(memcache: str) -> ThreadedMemcacheClient:
    host, port = parse_memcache_addr(memcache)
    return ThreadedMemcacheClient(host, port)



"""Our web services"""

import json
import re
from datetime import datetime

from feedgen.feed import FeedGenerator
from pyiem.util import utc
from twisted.internet import reactor
from twisted.python import log
from twisted.web import resource

import iembot.util as botutil
from iembot.slack import (
    SlackInstallChannel,
    SlackOauthChannel,
    SlackSubscribeChannel,
    SlackUnsubscribeChannel,
)
from iembot.types import JabberClient

XML_CACHE = {}
XML_CACHE_EXPIRES = {}


def wfo_rss(iembot: JabberClient, rm):
    """build a RSS for the given room"""
    if len(rm) == 4 and rm[0] == "k":
        rm = f"{rm[-3:]}chat"
    elif len(rm) == 3:
        rm = f"k{rm}chat"
    if rm not in iembot.chatlog:
        return ""
    if rm not in XML_CACHE:
        XML_CACHE[rm] = ""
        XML_CACHE_EXPIRES[rm] = -2
    # should not be empty given the caller
    lastID = iembot.chatlog[rm][0].seqnum
    if lastID == XML_CACHE_EXPIRES[rm]:
        return XML_CACHE[rm]

    rss = FeedGenerator()
    rss.generator("iembot")
    rss.title(f"{rm} IEMBot RSS Feed")
    rss.link(href=f"https://weather.im/iembot-rss/room/{rm}.xml", rel="self")
    rss.description(f"{rm} IEMBot RSS Feed")
    rss.lastBuildDate(f"{utc():%a, %d %b %Y %H:%M:%S} GMT")

    for entry in iembot.chatlog[rm]:
        botutil.add_entry_to_rss(entry, rss)

    XML_CACHE[rm] = rss.rss_str()
    XML_CACHE_EXPIRES[rm] = lastID
    return XML_CACHE[rm]


class RSSService(resource.Resource):
    """Our RSS service"""

    def isLeaf(self):
        """allow uri"""
        return True

    def __init__(self, iembot: JabberClient):
        """Constructor"""
        resource.Resource.__init__(self)
        self.iembot = iembot

    def render(self, request):
        if request.method == b"HEAD":
            return b""
        uri = request.uri.decode("utf-8")
        if uri.startswith("/wfo/"):
            tokens = re.findall(
                "/wfo/(k...|botstalk|...chat).xml", uri.lower()
            )
        else:
            tokens = re.findall("/room/(.*).xml", uri.lower())
        if not tokens:
            return b"ERROR!"

        rm = tokens[0]
        if uri.startswith("/wfo/"):
            if len(rm) == 4 and rm[0] == "k":
                rm = f"{rm[-3:]}chat"
            elif len(rm) == 3:
                rm = f"k{rm}chat"
        if not self.iembot.chatlog.get(rm, []):
            rss = FeedGenerator()
            rss.generator("iembot")
            rss.title("IEMBot RSS Feed")
            rss.link(
                href=f"https://weather.im/iembot-rss/room/{rm}.xml",
                rel="self",
            )
            rss.description("Syndication of iembot messages.")
            rss.lastBuildDate(f"{utc():%a, %d %b %Y %H:%M:%S} GMT")
            fe = rss.add_entry()
            fe.title("IEMBOT recently restarted, no history yet")
            fe.link(
                href="http://mesonet.agron.iastate.edu/projects/iembot/",
                rel="self",
            )
            fe.pubDate(utc().strftime("%a, %d %b %Y %H:%M:%S GMT"))
            xml = rss.rss_str()
        else:
            xml = wfo_rss(self.iembot, rm)
        request.setHeader("Content-Length", f"{len(xml)}")
        request.setHeader("Content-Type", "text/xml")
        request.setResponseCode(200)
        return xml


class RSSRootResource(resource.Resource):
    """I answer iembot-rss requests"""

    def __init__(self, iembot: JabberClient):
        """Constructor"""
        resource.Resource.__init__(self)
        service = RSSService(iembot)
        # legacy and lame
        self.putChild(b"wfo", service)
        # more properly alligned with what we do
        self.putChild(b"room", service)


# ------------------- iembot-json stuff below ---------------
class RoomChannel(resource.Resource):
    """respond to room requests"""

    def isLeaf(self):
        """allow uri calling"""
        return True

    def __init__(self, iembot: JabberClient):
        """Constructor"""
        resource.Resource.__init__(self)
        self.iembot = iembot

    def wrap(self, request, j):
        """Support specification of a JSONP callback"""
        if "callback" in request.args:
            request.setHeader("Content-type", "application/javascript")
            return (f"{request.args['callback'][0]}({j});").encode()
        return j.encode("utf-8")

    def render(self, request):
        """Process the request that we got, it should look something like:
        /room/dmxchat?seqnum=1
        """
        uri = request.uri.decode("utf-8")
        tokens = re.findall("/room/([a-z_0-9]+)", uri.lower())
        if not tokens:
            log.msg(f"Bad URI: {uri} len(tokens) is 0")
            return self.wrap(request, json.dumps("ERROR"))

        room = tokens[0]
        seqnum = request.args.get(b"seqnum")
        if seqnum is None or len(seqnum) != 1:
            log.msg(f"Bad URI: {request.uri} seqnum problem")
            return self.wrap(request, json.dumps("ERROR"))
        seqnum = int(seqnum[0])

        r = {"messages": []}
        for entry in self.iembot.chatlog.get(room, [])[::-1]:
            if entry.seqnum <= seqnum:
                continue
            ts = datetime.strptime(entry.timestamp, "%Y%m%d%H%M%S")
            r["messages"].append(
                {
                    "seqnum": entry.seqnum,
                    "ts": ts.strftime("%Y-%m-%d %H:%M:%S"),
                    "author": entry.author,
                    "product_id": entry.product_id,
                    "message": entry.log,
                }
            )

        return self.wrap(request, json.dumps(r))


class ReloadChannel(resource.Resource):
    """respond to /reload requests"""

    def isLeaf(self):
        """allow URI calling"""
        return True

    def __init__(self, iembot: JabberClient):
        """Constructor"""
        resource.Resource.__init__(self)
        self.iembot = iembot

    def render(self, _request):
        """Service the request."""
        log.msg("Reloading iembot room configuration....")
        self.iembot.reload_config(always_join=False)
        return json.dumps("OK").encode("utf-8")


class StatusChannel(resource.Resource):
    """respond to /status requests"""

    def __init__(self, iembot):
        """Constructor"""
        resource.Resource.__init__(self)
        self.iembot = iembot

    def render(self, _request):
        """Answer the call."""
        tp = reactor.getThreadPool()
        res = {
            "threadpool.max": tp.max,
            "threadpool.waiters": len(tp.waiters),
            "threadpool.working": len(tp.working),
        }
        return json.dumps(res).encode("utf-8")


class JSONRootResource(resource.Resource):
    """answer /iembot-json/ requests"""

    def __init__(self, iembot: JabberClient):
        """Constructor"""
        resource.Resource.__init__(self)
        self.putChild(b"room", RoomChannel(iembot))
        self.putChild(b"reload", ReloadChannel(iembot))
        self.putChild(b"status", StatusChannel(iembot))
        self.putChild(b"subscribe", SlackSubscribeChannel(iembot))
        self.putChild(b"unsubscribe", SlackUnsubscribeChannel(iembot))
        self.putChild(b"install", SlackInstallChannel(iembot))
        self.putChild(b"oauth_callback", SlackOauthChannel(iembot))



"""Utility functions for IEMBot"""

import copy
import glob
import os
import pickle
import pwd
import re
import socket
import traceback
from datetime import datetime, timedelta
from email.mime.text import MIMEText
from io import BytesIO
from zoneinfo import ZoneInfo

from pyiem.util import utc
from twisted.internet import reactor
from twisted.mail import smtp
from twisted.python import log
from twisted.words.xish import domish

import iembot
from iembot.types import JabberClient


def build_channel_subs(txn, auth_table: str) -> dict[int : list[str]]:
    """Build the subscriptions based on the given auth table."""
    # First find explicit subscriptions
    txn.execute(
        f"""
        select o.iembot_account_id, channel_name
        from {auth_table} o, iembot_subscriptions s, iembot_channels c
        WHERE o.iembot_account_id = s.iembot_account_id and s.channel_id = c.id
        """
    )
    subs = {}
    count_subs = 0
    for row in txn.fetchall():
        subs.setdefault(row["channel_name"], []).append(
            row["iembot_account_id"]
        )
        count_subs += 1
    # Find subscriptions based on groups
    txn.execute(
        """
        select o.iembot_account_id, channel_name from
        iembot_mastodon_oauth o, iembot_subscriptions s,
        iembot_channel_group_membership m, iembot_channels c
        WHERE o.iembot_account_id = s.iembot_account_id and
        s.group_id = m.group_id and m.channel_id = c.id
        """
    )
    for row in txn.fetchall():
        subs.setdefault(row["channel_name"], []).append(
            row["iembot_account_id"]
        )
        count_subs += 1
    log.msg(f"Built {count_subs} subscriptions from {auth_table}")
    return subs


def channels_room_list(bot: JabberClient, room: str):
    """
    Send a listing of channels that the room is subscribed to...
    @param room to list
    """
    channels = [
        channel
        for channel in bot.routingtable.keys()
        if room in bot.routingtable[channel]
    ]

    # Need to add a space in the channels listing so that the string does
    # not get so long that it causes chat clients to bail
    msg = f"This room is subscribed to {len(channels)} channels ({channels})"
    bot.send_groupchat(room, msg)


def channels_room_add(txn, bot: JabberClient, room: str, channel: str):
    """Add a channel subscription to a chatroom

    Args:
        txn (cursor): database transaction
        bot (JabberClient): bot instance
        room (str): the chatroom to add the subscription to
        channel (str): the channel to subscribe to for the room
    """
    # Remove extraneous fluff, all channels are uppercase
    channel = channel.upper().strip().replace(" ", "")
    if channel == "":
        bot.send_groupchat(
            room,
            "Failed to add channel to room subscription, you supplied a "
            "blank channel?",
        )
        return
    # Allow channels to be comma delimited
    for ch in channel.split(","):
        if ch not in bot.routingtable:
            bot.routingtable[ch] = []
        # If we are already subscribed, let em know!
        if room in bot.routingtable[ch]:
            bot.send_groupchat(
                room,
                "Error adding subscription, your room is already subscribed "
                f"to the '{ch}' channel",
            )
            continue
        # Add a channels entry for this channel, if one currently does
        # not exist
        txn.execute(
            "SELECT id from iembot_channels WHERE channel_name = %s",
            (ch,),
        )
        if txn.rowcount == 0:
            txn.execute(
                "INSERT into iembot_channels(channel_name, description) "
                "VALUES (%s, %s) returning id",
                (ch, ch),
            )
        channel_id = txn.fetchone()["id"]

        # Add to routing table
        bot.routingtable[ch].append(room)
        # Add to database
        txn.execute(
            """
    INSERT into iembot_subscriptions (iembot_account_id, channel_id)
    values (
        (select iembot_account_id from iembot_rooms where roomname = %s),
        %s
    )
    """,
            (room, channel_id),
        )
        bot.send_groupchat(room, f"Subscribed {room} to channel '{ch}'")
    # Send room a listing of channels!
    channels_room_list(bot, room)


def channels_room_del(txn, bot: JabberClient, room: str, channel: str):
    """Removes a channel subscription for a given room

    Args:
        txn (cursor): database cursor
        room (str): room to unsubscribe
        channel (str): channel to unsubscribe from
    """
    channel = channel.upper().strip().replace(" ", "")
    if channel == "":
        bot.send_groupchat(room, "Blank or missing channel")
        return

    for ch in channel.split(","):
        if ch not in bot.routingtable:
            bot.send_groupchat(room, f"Unknown channel: '{ch}'")
            continue

        if room not in bot.routingtable[ch]:
            bot.send_groupchat(room, f"Room not subscribed to channel: '{ch}'")
            continue

        # Remove from routing table
        bot.routingtable[ch].remove(room)
        # Remove from database
        txn.execute(
            """
    DELETE from iembot_subscriptions WHERE
    iembot_account_id = (
    select iembot_account_id from iembot_rooms where roomname = %s) and
    channel_id = (
    select id from iembot_channels where channel_name = %s)
    """,
            (room, ch),
        )
        bot.send_groupchat(room, f"Unsubscribed {room} to channel '{ch}'")
    channels_room_list(bot, room)


def purge_logs(bot: JabberClient):
    """Remove chat logs on a 24 HR basis"""
    log.msg("purge_logs() called...")
    basets = utc() - timedelta(
        days=int(bot.config.get("bot.purge_xmllog_days", 7))
    )
    for fn in glob.glob("logs/xmllog.*"):
        ts = datetime.strptime(fn, "logs/xmllog.%Y_%m_%d")
        ts = ts.replace(tzinfo=ZoneInfo("UTC"))
        if ts < basets:
            log.msg(f"Purging logfile {fn}")
            os.remove(fn)


def email_error(exp, bot: JabberClient, message=""):
    """
    Something to email errors when something fails
    """
    # Always log a message about our fun
    cstr = BytesIO()
    if isinstance(exp, Exception):
        traceback.print_exc(file=cstr)
        cstr.seek(0)
        if isinstance(exp, Exception):
            log.err(exp)
        else:
            log.msg(exp)
    log.msg(message)

    def should_email():
        """Should we send an email?"""
        # bot.email_timestamps contains timestamps of emails we *sent*
        utcnow = utc()
        # If we don't have any entries, we should email!
        if len(bot.email_timestamps) < 10:
            bot.email_timestamps.insert(0, utcnow)
            return True
        delta = utcnow - bot.email_timestamps[-1]
        # Effectively limits to 10 per hour
        if delta < timedelta(hours=1):
            return False
        # We are going to email!
        bot.email_timestamps.insert(0, utcnow)
        # trim listing to 10 entries
        while len(bot.email_timestamps) > 10:
            bot.email_timestamps.pop()
        return True

    # Logic to prevent email bombs
    if not should_email():
        log.msg("Email threshold exceeded, so no email sent!")
        return False

    le = " ".join([f"{_:.2f}" for _ in os.getloadavg()])
    tb = cstr.getvalue().decode("utf-8")
    if tb != "":
        tb = f"Exception       : {tb}\n"
    expmsg = ""
    if exp is not None:
        expmsg = f"Exception       : {exp}\n"
    msg = MIMEText(
        f"System          : {pwd.getpwuid(os.getuid())[0]}@"
        f"{socket.gethostname()} [CWD: {os.getcwd()}]\n"
        f"System UTC date : {utc()}\n"
        f"process id      : {os.getpid()}\n"
        f"iembot.version  : {iembot.__version__}\n"
        f"system load     : {le}\n"
        f"{tb}"
        f"{expmsg}"
        f"Message: {message}\n"
    )
    msg["subject"] = f"[bot] Traceback -- {socket.gethostname()}"

    msg["From"] = bot.config.get("bot.email_errors_from", "root@localhost")
    msg["To"] = bot.config.get("bot.email_errors_to", "root@localhost")

    df = smtp.sendmail(
        bot.config.get("bot.smtp_server", "localhost"),
        msg["From"],
        msg["To"],
        msg,
    )
    df.addErrback(log.err)
    return True


def load_chatrooms_from_db(txn, bot: JabberClient, always_join: bool = False):
    """Load database configuration and do work

    Args:
      txn (dbtransaction): database cursor
      bot (JabberClient): the running bot instance
      always_join (boolean): do we force joining each room, regardless
    """
    # Load up a list of chatrooms
    txn.execute(
        "SELECT iembot_account_id, roomname from iembot_rooms "
        "WHERE roomname is not null ORDER by roomname ASC"
    )
    oldrooms = list(bot.rooms.keys())
    joined = 0
    xref = {}
    if always_join or "botstalk" not in oldrooms:
        # botstalk is special and should be joined immediately
        presence = domish.Element(("jabber:client", "presence"))
        presence["to"] = f"botstalk@{bot.conference}/{bot.myjid.user}"
        bot.xmlstream.send(presence)

    for i, row in enumerate(txn.fetchall()):
        rm = row["roomname"]
        xref[row["iembot_account_id"]] = rm
        # Setup Room Config Dictionary
        if rm not in bot.rooms:
            bot.rooms[rm] = {
                "occupants": {},
                "joined": False,
            }

        if always_join or rm not in oldrooms:
            presence = domish.Element(("jabber:client", "presence"))
            presence["to"] = f"{rm}@{bot.conference}/{bot.myjid.user}"
            # Some jitter to prevent overloading
            reactor.callLater(i % 30, bot.xmlstream.send, presence)
            joined += 1
        if rm in oldrooms:
            oldrooms.remove(rm)

    # Check old rooms for any rooms we need to vacate!
    for rm in oldrooms:
        presence = domish.Element(("jabber:client", "presence"))
        presence["to"] = f"{rm}@{bot.conference}/{bot.myjid.user}"
        presence["type"] = "unavailable"
        bot.xmlstream.send(presence)

        del bot.rooms[rm]
    log.msg(
        f"... loaded {txn.rowcount} chatrooms, joined {joined} of them, "
        f"left {len(oldrooms)} of them"
    )

    # Doing an ugly pivot with this
    rt_using_account_ids = build_channel_subs(
        txn,
        "iembot_rooms",
    )
    rt = {}
    for channel, accounts in rt_using_account_ids.items():
        rt[channel] = [xref[account] for account in accounts]
    bot.routingtable = rt


def load_webhooks_from_db(txn, bot: JabberClient):
    """Load twitter config from database"""
    txn.execute(
        "SELECT channel, url from iembot_webhooks "
        "WHERE channel is not null and url is not null"
    )
    table = {}
    for row in txn.fetchall():
        url = row["url"]
        channel = row["channel"]
        if url == "" or channel == "":
            continue
        res = table.setdefault(channel, [])
        res.append(url)
    bot.webhooks_routingtable = table
    log.msg(f"load_webhooks_from_db(): {txn.rowcount} subs found")


def load_chatlog(bot: JabberClient):
    """load up our pickled chatlog"""
    if not os.path.isfile(bot.picklefile):
        log.msg(f"pickfile not found: {bot.picklefile}")
        return
    try:
        with open(bot.picklefile, "rb") as fh:
            oldlog = pickle.load(fh)
        for rm in oldlog:
            rmlog = oldlog[rm]
            bot.chatlog[rm] = copy.deepcopy(rmlog)
            if not rmlog:
                continue
            # First message in list is the newest :/
            seq = rmlog[0].seqnum
            if seq is not None and int(seq) > bot.seqnum:
                bot.seqnum = int(seq)
        log.msg(
            f"Loaded CHATLOG pickle: {bot.picklefile}, seqnum: {bot.seqnum}"
        )
    except Exception as exp:
        log.err(exp)


def htmlentities(text):
    """Escape chars in the text for HTML presentation

    Args:
      text (str): subject to replace

    Returns:
      str : result of replacement
    """
    for lookfor, replacewith in [
        ("&", "&amp;"),
        (">", "&gt;"),
        ("<", "&lt;"),
        ("'", "&#39;"),
        ('"', "&quot;"),
    ]:
        text = text.replace(lookfor, replacewith)
    return text


def remove_control_characters(html):
    """Get rid of cruft?"""
    # https://github.com/html5lib/html5lib-python/issues/96
    html = re.sub("[\x00-\x08\x0b\x0e-\x1f\x7f]", "", html)
    return html


def add_entry_to_rss(entry, rss):
    """Convert a txt Jabber room message to a RSS feed entry

    Args:
      entry(CHAT_LOG_ENTRY): entry

    Returns:
      PyRSSGen.RSSItem
    """
    ts = datetime.strptime(entry.timestamp, "%Y%m%d%H%M%S")
    txt = entry.txtlog
    m = re.search(r"https?://", txt)
    urlpos = -1
    if m:
        urlpos = m.start()
    else:
        txt += "  "
    ltxt = txt[urlpos:].replace("&amp;", "&").strip()
    if ltxt == "":
        ltxt = "https://mesonet.agron.iastate.edu/projects/iembot/"
    fe = rss.add_entry(order="append")
    fe.title(txt[:urlpos].strip())
    fe.link(link={"href": ltxt})
    txt = remove_control_characters(entry.product_text)
    fe.content(f"<pre>{htmlentities(txt)}</pre>", type="CDATA")
    fe.pubDate(ts.strftime("%a, %d %b %Y %H:%M:%S GMT"))


def daily_timestamp(bot: JabberClient):
    """Send a timestamp to each room we are in.

    Args:
      bot (JabberClient) instance
    """
    # Make sure we are a bit into the future!
    utc0z = utc() + timedelta(hours=1)
    utc0z = utc0z.replace(hour=0, minute=0, second=0, microsecond=0)
    mess = f"------ {utc0z:%b %-d, %Y} [UTC] ------"
    for rm in bot.rooms:
        bot.send_groupchat(rm, mess)

    tnext = utc0z + timedelta(hours=24)
    delta = (tnext - utc()).total_seconds()
    log.msg(f"Calling daily_timestamp in {delta:.2f} seconds")
    return reactor.callLater(delta, daily_timestamp, bot)



"""Send content to various webhooks."""

import json
from io import BytesIO

from twisted.internet import reactor
from twisted.python import log
from twisted.web.client import Agent, FileBodyProducer, readBody
from twisted.web.http_headers import Headers

from iembot.types import JabberClient


def route(bot: JabberClient, channels, elem):
    """Route messages found in provided elem.

    Args:
      bot: iembot instance.
      channels (list): channels for this message.
      elem: xish element.
    """
    subs = [
        bot.webhooks_routingtable[channel]
        for channel in channels
        if channel in bot.webhooks_routingtable
    ]
    if not subs:
        return
    data = {"text": str(elem.body)}
    postdata = json.dumps(data).encode("utf-8", "ignore")
    used = []
    for hooks in subs:
        for hook in hooks:
            if hook in used:
                continue
            used.append(hook)
            bp = FileBodyProducer(BytesIO(postdata))
            defer = Agent(reactor).request(
                method=b"POST",
                uri=hook.encode("ascii"),
                headers=Headers({"Content-type": ["application/json"]}),
                bodyProducer=bp,
            )
            defer.addCallback(_cb)
            defer.addErrback(_eb)


def _eb(*args):
    """errback."""
    log.msg(str(args))


def _cb(*args):
    d = readBody(args[0])
    d.addCallback(_cbBody)
    return d


def _cbBody(body):
    """deferred deferred."""
    log.msg(body)
